########################################################################
## SPINN DESIGN CODE
# YOUTUBE: (SPINN TV) https://www.youtube.com/spinnTv
# WEBSITE: spinncode.com
########################################################################
from qtpy.QtCore import Qt, QEasingCurve, QPropertyAnimation, QSize, Property, QPoint, QEnum, QObject
from qtpy.QtGui import QPalette, QIcon, QPaintEvent, QPainter, QColor
from qtpy.QtWidgets import QCheckBox, QApplication, QLabel, QStyleOption, QStyle

import os

class QCustomCheckBox(QCheckBox):
    script_dir = os.path.dirname(os.path.realpath(__file__))
    WIDGET_ICON = os.path.join(script_dir, "components/icons/check_box.png")
    WIDGET_TOOLTIP = "A custom animated checkbox widget"
    WIDGET_DOM_XML = """
    <ui language='c++'>
        <widget class='QCustomCheckBox' name='customCheckBox'>
        </widget>
    </ui>
    """
    WIDGET_MODULE="Custom_Widgets.QCustomCheckBox"
    
    def __init__(self, parent=None):
        super().__init__(parent)

        self.setCursor(Qt.PointingHandCursor)

        # Check if a QApplication instance already exists
        if QApplication.instance():
            app = QApplication.instance()
        else:
            app = QApplication([])


        # Get the default palette
        palette = app.palette()
            
        # Get the default background color
        bgColor = palette.color(QPalette.Window)
        # Get the default text color (assuming this is your "circle_color")
        circleColor = palette.color(QPalette.Text)
        # Get the default highlight color (assuming this is your "active_color")
        activeColor = palette.color(QPalette.Highlight)

        # COLORS
        self.bgColor = bgColor
        self._circleColor = circleColor
        self._activeColor = activeColor

        # Animation
        self._animationEasingCurve = QEasingCurve.OutBounce
        self._easing_curve = 0
        self._animationDuration = 300

        self.pos = 0
        self.animation = QPropertyAnimation(self, b"position")
        self.animation.setEasingCurve(self._animationEasingCurve)
        self.animation.setDuration(self._animationDuration)
        self.stateChanged.connect(self.setup_animation)

        # Create a QLabel to display the text
        self.label = QLabel(self)
        # self.label.setContentsMargins(0, 0, 0, 0)  # Set contents margins to 0 to remove spacing
        self.label.setWordWrap(True)

        # Initialize icon
        self.icon = QIcon()
        self._iconSize = QSize(0, 0)  # Default icon size

    @Property(QColor)
    def backgroundColor(self):
        return self.bgColor
    
    @backgroundColor.setter
    def backgroundColor(self, color):
        self.bgColor = color
    
    @Property(QColor)
    def circleColor(self):
        return self._circleColor
    
    @circleColor.setter
    def circleColor(self, color):
        self._circleColor = color

    @Property(QColor)
    def activeColor(self):
        return self._activeColor
    
    @activeColor.setter
    def activeColor(self, color):
        self._activeColor = color

    @Property(int)
    def animationDuration(self):
        return self._animationDuration
    
    @animationDuration.setter
    def animationDuration(self, duration):
        self._animationDuration = duration

    @Property(QEasingCurve, designable=True)
    def animationEasingCurve(self):
        return self._easing_curve
    
    @animationEasingCurve.setter
    def animationEasingCurve(self, curve):
        # if not isinstance(curve, EasingCurveEnum):
        #     raise ValueError("Invalid easing curve enum")
            
        self._easing_curve = curve

    def setIcon(self, icon):
        self.icon = icon
        self.update()
    
    def setIconSize(self, size):
        """
        Set the size of the icon for the checkbox.

        Parameters:
            size (QSize): The size of the icon.
        """
        self._iconSize = size
        self.update()

    ########################################################################
    # Customize QCustomCheckBox
    ########################################################################
    def customizeQCustomCheckBox(self, **customValues):
        if "bgColor" in customValues:
            self.bgColor = customValues["bgColor"]
        
        if "circleColor" in customValues:
            self._circleColor = customValues["circleColor"]

        if "activeColor" in customValues:
            self._activeColor = customValues["activeColor"]

        if "animationEasingCurve" in customValues:
            self.animationEasingCurve = customValues["animationEasingCurve"]
            self.animation.setEasingCurve(self.animationEasingCurve)

        if "animationDuration" in customValues:
            self._animationDuration = customValues["animationDuration"]
            self.animation.setDuration(self._animationDuration)

        self.update()
    
    def showEvent(self, e):
        super().showEvent(e)
        self.adjustWidgetSize()
        self.update()


    def resizeEvent(self, event):
        super().resizeEvent(event)
        self.adjustWidgetSize()

    def adjustWidgetSize(self):
        # self.adjustSize()
        # Update checkbox size
        # Update label position and width
        icon_size = self._iconSize.width()
        label_margin = 5  # Adjust the margin between the icon and the label

        if not self.label.text().strip():
            self.label.setFixedSize(0, 0)
        else:
            label_width = self.width() - (self.height() * 2.1 + icon_size + label_margin)  # Calculate the width of the label area
            label_height = self.height()  # Use the height of the checkbox for the label height

            self.label.adjustSize()
            
            label_x = self.height() * 2.1 + icon_size + label_margin  # Calculate the x position of the label
            label_y = (self.height() - label_height) / 2  # Center the label vertically within the checkbox area

            self.label.setGeometry(label_x, label_y, label_width, label_height)
        
        self.update()


    def setText(self, text):
        # super().setText(text)
        self.label.setText(text)
        self.adjustWidgetSize()

    @Property(float, designable=False)
    def position(self):
        return self.pos
    
    @position.setter
    def position(self, pos):
        self.pos = pos
        self.update()

    # START STOP ANIMATION
    def setup_animation(self, value):
        self.pos = 0
        self.animation.stop()
        if value:
            self.animation.setEndValue(self.height() + 2)
        else:
            self.animation.setEndValue(0)
        self.animation.start()
    
    def hitButton(self, pos: QPoint):
        return self.contentsRect().contains(pos)

    def paintEvent(self, e: QPaintEvent):
        super().paintEvent(e)
     
        opt = QStyleOption()
        opt.initFrom(self)
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        self.style().drawPrimitive(QStyle.PE_Widget, opt, painter, self)

        # SET PEN
        painter.setPen(Qt.NoPen)

        # Define margins
        margin = 3

        # DRAW ICON (Optional)
        if not self.icon.isNull():
            icon_size = self.height() * 0.7 if self.icon.availableSizes() else QSize(16, 16)  
            if self._iconSize == QSize(0, 0):
                self._iconSize = QSize(icon_size, icon_size)
            
            pixmap = self.icon.pixmap(self._iconSize)
            # Adjust horizontal position to add margin between checkbox and icon
            icon_x = self.height() * 2.1 + margin
            icon_y = (self.height() - self._iconSize.height()) / 2  # Center the icon vertically within the checkbox area
            painter.drawPixmap(icon_x, icon_y, pixmap)

        if not self.isChecked():
            # Draw rounded rectangle for unchecked state
            painter.setBrush(QColor(self.bgColor))
            painter.drawRoundedRect(0, 0, self.height() * 2.1, self.height(), self.height() * .5, self.height() * .5)

            # Draw circle for unchecked state
            painter.setBrush(QColor(self._circleColor))
            painter.drawEllipse(self.pos, 0, self.height(), self.height())
        else:
            # Draw rounded rectangle for checked state
            painter.setBrush(QColor(self._activeColor))
            painter.drawRoundedRect(0, 0, self.height() * 2.1, self.height(), self.height() * .5, self.height() * .5)

            # Draw circle for checked state
            painter.setBrush(QColor(self._circleColor))
            painter.drawEllipse(self.pos, 0, self.height(), self.height())
        
        self.adjustWidgetSize()

        painter.end()
        



#                 PyQt5 Custom Widgets                #
#                GPL 3.0 - Kadir Aksoy                #
#   https://github.com/kadir014/pyqt5-custom-widgets  #


import time
from math import sin, cos, radians
from qtpy.QtCore import Qt, QRectF
from qtpy.QtWidgets import QWidget
from qtpy.QtGui import QPainter, QPen, QPalette, QColor

class QCustomSpinner(QWidget):
    def __init__(self, lineWidth = 2, lineColor = None, direction = "Clockwise", borderRadius = 3, animationType = "Bounce"):
        super().__init__()

        self.w = lineWidth
        if lineColor is None:
            self.color = self.palette().color(QPalette.Highlight)
        else:
            self.color = lineColor
        self.direction = direction
        self.borderRadius = borderRadius

        self.angle = 0
        self.speed = 4.8

        self.animType = animationType

        self.play = True

        self.last_call = time.time()

    def __repr__(self):
        return f"<QCustom.Spinner()>"

    def paintEvent(self, event):
        pt = QPainter()
        pt.begin(self)
        pt.setRenderHint(QPainter.Antialiasing, on=True)

        w = self.w
        try:
            pen = QPen(self.color, w)
        except:
            pen = QPen(QColor(self.color), w)
        # Set the line cap style to round
        pen.setCapStyle(Qt.RoundCap)
        pt.setPen(pen)

        if self.direction == "Counterclockwise":
            start_angle = self.angle
            end_angle = self.angle + 90 * 16
        else:  # 
            start_angle = self.angle
            end_angle = self.angle - 90 * 16

        if self.animType == "Smooth":
            pt.drawArc(w, w, self.width() - w * 2, self.height() - w * 2, start_angle, 90 * 16)
        elif self.animType == "Bounce":
            sa = ((sin(radians(start_angle / 16)) + 1) / 2) * (180 * 16) + (
                        (sin(radians((end_angle / 16) + 130)) + 1) / 2) * (180 * 16)
            pt.drawArc(w, w, self.width() - w * 2, self.height() - w * 2, start_angle, sa)

        pt.end()

        ep = (time.time() - self.last_call) * 1000
        self.last_call = time.time()

        self.angle += self.speed * ep
        if self.angle > 360 * 16:
            self.angle = 0
        elif self.angle < 0:
            self.angle = 360 * 16

        if self.play:
            self.update()





from qtpy.QtWidgets import QComboBox, QMainWindow
from qtpy.QtCore import Qt, Signal, Property

from Custom_Widgets.QCustomTheme import QCustomTheme
from Custom_Widgets.QAppSettings import QAppSettings

import os

class QCustomThemeList(QComboBox):
    # Icon path for the widget
    script_dir = os.path.dirname(os.path.realpath(__file__))
    WIDGET_ICON = os.path.join(script_dir, "components/icons/palette.png")
    
    # Tooltip for the widget
    WIDGET_TOOLTIP = "A custom QComboBox for selecting themes"
    
    # XML string for the widget
    WIDGET_DOM_XML = """
    <ui language='c++'>
        <widget class="QCustomThemeList" name="CustomThemeList">
        <property name="windowTitle">
        <string>Custom Theme List</string>
        </property>
        </widget>
    </ui>
    """

    WIDGET_MODULE="Custom_Widgets.QCustomThemeList"

    themeChanged = Signal(str)  # Custom signal for theme change

    def __init__(self, parent = None, *args, **kwargs):
        # super().__init__(*args, **kwargs)
        super(QCustomThemeList, self).__init__(parent)

        self.setDuplicatesEnabled(False)

        self._loadPredefinedThemes = False  # Default value
        self.themeEngine = QCustomTheme()
        self._app_themes = self.themeEngine.themes
        
        # Get initial themes based on loadPredefinedThemes setting
        new_theme_names = self.get_filtered_themes()
        self.populate_themes(new_theme_names)

        self.currentIndexChanged.connect(self.on_theme_changed)

        self.old_theme_names = []
        self.new_theme_names = []

    @Property(bool)
    def loadPredefinedThemes(self):
        """Get the loadPredefinedThemes property value"""
        return self._loadPredefinedThemes

    @loadPredefinedThemes.setter
    def loadPredefinedThemes(self, value):
        """Set the loadPredefinedThemes property value and refresh themes"""
        if self._loadPredefinedThemes != value:
            self._loadPredefinedThemes = value
            # Refresh the theme list when this property changes
            new_theme_names = self.get_filtered_themes()
            self.populate_themes(new_theme_names)

    def get_filtered_themes(self):
        """Get themes filtered based on loadPredefinedThemes setting"""
        if not self._loadPredefinedThemes:
            # Filter out predefined themes (where theme.predefined is True)
            filtered_themes = [theme.name for theme in self.themeEngine.themes 
                            if not theme.predefined]
            return sorted(filtered_themes)
        else:
            # Return all theme names
            all_theme_names = [theme.name for theme in self.themeEngine.themes]
            return sorted(all_theme_names)

    def populate_themes(self, new_theme_names):
        try:
            self.blockSignals(True)
            self.clear()
            for theme in new_theme_names:
                self.addItem(theme)
                
                if theme == self.themeEngine.theme:
                    # Select the matching theme
                    self.setCurrentText(theme)
            
            self.blockSignals(False)
                    
        except Exception as e:
            # print(e)
            pass

    def on_theme_changed(self, index):
        selected_theme = self.currentText()
        self.themeEngine.theme = selected_theme
        self.themeChanged.emit(selected_theme)  

    def check_theme_updates(self):
        # Use maximum available size
        self.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        
        try:
            self.themeEngine = QCustomTheme()
            self._app_themes = self.themeEngine.themes

            # Get filtered themes based on loadPredefinedThemes setting
            self.new_theme_names = self.get_filtered_themes()
            
            # Compare the sorted lists of names
            if self.old_theme_names != self.new_theme_names:
                self.populate_themes(self.new_theme_names)
                self.old_theme_names = self.new_theme_names
                
        except Exception as e:
            print(f"Error: {e}")

    def showEvent(self, event):
        """Handle show event to adjust size and refresh themes"""
        super().showEvent(event)
        self.adjustSize()
        self.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        self.check_theme_updates()

    def resizeEvent(self, event):
        """Handle resize event to use maximum available size"""
        super().resizeEvent(event)
        self.setSizeAdjustPolicy(QComboBox.AdjustToContents)

    def paintEvent(self, event):
        super(QCustomThemeList, self).paintEvent(event)
        
        self.check_theme_updates()


from qtpy.QtGui import *
from qtpy.QtWidgets import *

def setNewIcon(self, url):
    icon = QIcon(url)
    self.setIcon(icon)

    self.iconUrl = url

def setNewPixmap(self, url):
    piximap = QPixmap(url)
    self.setPixmap(piximap)

    self.piximapUrl = url

def setNewTabIcon(self, tabName, url):
    icon = QIcon(url)

    # Find the index of the tab with the name "Tab 2"
    tab_index = -1
    for index in range(self.count()):
        tab_object_name = self.widget(index).objectName()
        if tab_object_name == tabName:
            tab_index = index
            break
    # Change icon of the tab with the name "Tab 2"
    if tab_index != -1:
        self.setTabIcon(tab_index, icon)

    self.iconUrl = url

def setNewToolBoxIcon(self, itemName, url):
    icon = QIcon(url)

    # Find the index of the item with the specified name
    item_index = -1
    for index in range(self.count()):
        item_text = self.widget(index).objectName()
        if item_text == itemName:
            item_index = index
            break

    # Change icon of the item with the specified name
    if item_index != -1:
        self.setItemIcon(item_index, icon)

    self.iconUrl = url

    



import sys
import os
import importlib.util
from PySide6.QtGui import QResizeEvent
from qtpy.QtWidgets import QWidget, QStyleOption, QStyle, QLabel, QVBoxLayout, QSizePolicy
from qtpy.QtGui import QPainter
from qtpy.QtCore import Property, Qt

from Custom_Widgets.QCustomTheme import QCustomTheme
from Custom_Widgets.Utils import is_in_designer
from Custom_Widgets.QCustomComponentLoader import QCustomComponentLoader

class QCustomComponentContainer(QWidget):
    """A custom widget to load and display a UI class defined in an external file."""

    script_dir = os.path.dirname(os.path.realpath(__file__))
    WIDGET_ICON = os.path.join(script_dir, "components/icons/view_quilt.png")
    WIDGET_TOOLTIP = "A custom component loader for dynamic UI loading."
    WIDGET_DOM_XML = """
    <ui language='c++'>
        <widget class="QCustomComponentContainer" name="QCustomComponentContainer">
        </widget>
    </ui>
    """
    
    WIDGET_MODULE = "Custom_Widgets.QCustomComponentContainer"

    def __init__(self, parent=None):
        super().__init__(parent)
        self._layout = None

        # Initialize UI class and setup
        self._ui_class = None
        self._file_path = None
        self._form_class = None
        self.ui = None

        self._designer_preview = False
        self._is_designer_mode = False
        self.form = QCustomComponentLoader()
    
    def showEvent(self, e):
        super().showEvent(e)
        # Use a single shot timer to avoid recursive layout issues
        if self._form_class and self._file_path and not hasattr(self, "component"):
            from qtpy.QtCore import QTimer
            QTimer.singleShot(0, self._refresh_component)

    def _refresh_component(self):
        # Clear any existing layout and labels
        if self.layout() is not None:
            QWidget().setLayout(self.layout())  # Reset layout
        
        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self.setLayout(self._layout)

        self.form = QCustomComponentLoader()
        self.form.previewComponent = self.previewComponent
        self.form.loadComponent(formClassName=self._form_class, filePath=self._file_path)

        self.layout().addWidget(self.form) 
        
        try:
            #older versions
            self.form.form =  self.form.ui 
            self.shownForm =  self.form.ui  

            # for components
            self.component =  self.form.ui

        except:
            self.shownForm = None

    @Property(str)
    def filePath(self):
        """Property to get or set the file path of the UI class."""
        return self._file_path

    @filePath.setter
    def filePath(self, value: str):
        if self._file_path != value:
            self._file_path = os.path.normpath(value)

            self._refresh_component()

    @Property(str)
    def formClassName(self):
        """Property to get or set the form class name."""
        return self._form_class.__name__ if self._form_class else ""

    @formClassName.setter
    def formClassName(self, value: str):
        if self._form_class != value:
            self._form_class = value

            self._refresh_component()

    @Property(bool)
    def previewComponent(self):
        """Property to get or set the form class name."""
        return self._designer_preview

    @previewComponent.setter
    def previewComponent(self, value: bool):
        if self._designer_preview != value:
            self._designer_preview = value

            self._refresh_component()

    def paintEvent(self, e):
        """Handle the paint event to customize the appearance of the widget."""
        super().paintEvent(e)
        opt = QStyleOption()
        opt.initFrom(self)
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        self.style().drawPrimitive(QStyle.PE_Widget, opt, painter, self)
    
    def resizeEvent(self, event: QResizeEvent) -> None:  
        # self.adjustSize()

        return super().resizeEvent(event)
    




from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *

import os


class ProgressBarArcLoader:
    def __init__(self, parent=None, progress=0, duration=2000):
        self.parent = parent
        self.progress = progress  # Current progress percentage
        self.duration = duration  # Animation duration in milliseconds

    def getProgressAnimation(self, start, end):
        animation = QVariantAnimation(self.parent)
        animation.setStartValue(start)
        animation.setEndValue(end)
        animation.setDuration(self.duration)
        animation.setEasingCurve(QEasingCurve.InOutSine)
        animation.valueChanged.connect(self.updateProgress)
        return animation

    def updateProgress(self, newValue):
        self.progress = newValue
        self.parent.update()

class QCustomRoundProgressBar(QWidget):
    # Define XML for Qt Designer
    script_dir = os.path.dirname(os.path.realpath(__file__))
    WIDGET_ICON = os.path.join(script_dir, "../components/icons/donut_large.png")
    WIDGET_TOOLTIP = "A custom round progress bar that shows animated progress."
    WIDGET_DOM_XML = """
    <ui language='c++'>
        <widget class='QCustomRoundProgressBar' name='customRoundProgressBar'>
        </widget>
    </ui>
    """
    WIDGET_MODULE = "Custom_Widgets.QCustomProgressBars"

    def __init__(self, parent=None, progressColor=None, textColor=None, progressBarWidth=5):
        super().__init__(parent)

        # Set default colors from application palette if not specified
        appPalette = QApplication.palette()
        self._progressColor = progressColor if progressColor is not None else appPalette.highlight().color()  # Accent color
        self._progressBaseColor = QColor(255, 255, 255, 100)  # Translucent white
        self._textColor = textColor if textColor is not None else appPalette.text().color()  # Default text color

        self._min = 0
        self._max = 100
        self._value = 0
        self._textVisible = True
        self._clockwise = True  # Direction of progress bar (True for clockwise)
        self._animationDuration = 500  # Default animation duration
        self._progressBarWidth = progressBarWidth  # Store the progress bar width
        self.initPen(self._progressBarWidth)

        # Initialize the loader
        self.arcLoader = ProgressBarArcLoader(self, 0, self._animationDuration)
        self.currentProgress = 0

        # Set size policy to make it responsive
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setMinimumSize(160, 160)

        # Start the initial progress animation to 0
        self.animateTo(0)

    # Set up properties to be editable in Qt Designer
    @Property(int)
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        self._value = val
        self.animateTo(val)

    def setValue(self, val):
        self.value = val

    @Property(int)
    def minimum(self):
        return self._min

    @minimum.setter
    def minimum(self, val):
        self._min = val
        self.update()

    @Property(int)
    def maximum(self):
        return self._max

    @maximum.setter
    def maximum(self, val):
        self._max = val
        self.update()

    @Property(bool)
    def textVisible(self):
        return self._textVisible

    @textVisible.setter
    def textVisible(self, visible):
        self._textVisible = visible
        self.update()

    @Property(bool)
    def clockwise(self):
        return self._clockwise

    @clockwise.setter
    def clockwise(self, val):
        self._clockwise = val
        self.update()

    @Property(int)
    def animationDuration(self):
        return self._animationDuration

    @animationDuration.setter
    def animationDuration(self, duration):
        self._animationDuration = duration
        self.arcLoader.duration = duration  # Update the animation loader's duration
        self.update()

    @Property(int)
    def progressBarWidth(self):
        return self._progressBarWidth

    @progressBarWidth.setter
    def progressBarWidth(self, width):
        self._progressBarWidth = width
        self.initPen(width)  # Reinitialize the pen with the new width
        self.update()  # Update the widget to reflect changes

    @Property(QColor)
    def progressColor(self):
        return self._progressColor

    @progressColor.setter
    def progressColor(self, color):
        self._progressColor = color
        self.initPen(self.pen.width())  # Re-initialize the pen with the new color
        self.update()

    @Property(QColor)
    def progressBaseColor(self):
        return self._progressBaseColor

    @progressBaseColor.setter
    def progressBaseColor(self, color):
        self._progressBaseColor = color
        self.initPen(self.pen.width())  # Re-initialize the pen with the new color
        self.update()

    @Property(QColor)
    def textColor(self):
        return self._textColor

    @textColor.setter
    def textColor(self, color):
        self._textColor = color
        self.update()

    def initPen(self, progressBarWidth):
        self.pen = QPen()
        self.pen.setColor(self._progressColor)  # Use the progress bar color property
        self.pen.setWidth(progressBarWidth)
        self.pen.setCapStyle(Qt.RoundCap)

    def animateTo(self, progress):
        """Animates the progress bar to the specified progress percentage."""
        startProgress = (self.arcLoader.progress / self.maximum) * 100
        endProgress = (progress / self.maximum) * 100
        animation = self.arcLoader.getProgressAnimation(startProgress, endProgress)
        self.currentProgress = progress
        animation.start()

    def calculateXR(self):
        x = self.pen.width() / 2
        r = min(self.width(), self.height()) - self.pen.width()
        return x, r

    def draw(self):
        # Calculate dimensions based on widget size
        x, r = self.calculateXR()
        progressArc = int((self.arcLoader.progress / 100) * 360)

        # Draw the base arc (translucent white background for the progress path)
        basePen = QPen(self._progressBaseColor) 
        basePen.setWidth(self.pen.width())
        basePen.setCapStyle(Qt.RoundCap)
        self.painter.setPen(basePen)
        self.painter.drawArc(x, x, r, r, 0, 360 * 16)  # Full circle for the base

        # Choose the direction of the arc based on the 'clockwise' property
        self.painter.setPen(self.pen)  # Use the progress bar pen
        if self._clockwise:
            self.painter.drawArc(x, x, r, r, 90 * 16, -progressArc * 16)  # Clockwise progress arc
        else:
            self.painter.drawArc(x, x, r, r, 90 * 16, progressArc * 16)  # Counterclockwise progress arc

        # Draw the progress text in the center
        if self.textVisible:
            self.drawText()


    def drawText(self):
        # Set the text font and color
        font = self.font()  # Adjust font size based on widget size
        self.painter.setFont(font)
        self.painter.setPen(QPen(self._textColor))  # Use the text color property
        text = f"{int(self.arcLoader.progress)}%"  # Progress percentage

        # Calculate the bounding rectangle to center the text
        textRect = self.rect()
        self.painter.drawText(textRect, Qt.AlignCenter, text)

    def paintEvent(self, e):
        self.painter = QPainter(self)
        self.painter.setRenderHint(QPainter.Antialiasing)
        self.painter.setPen(self.pen)
        self.draw()
        self.painter.end()



from qtpy import QtGui, QtCore, QtWidgets
import typing
    
class QTagEdit(QtWidgets.QScrollArea):
    """A tag based edit"""

    def __init__(self, parent: QtWidgets = None, tag_suggestions: typing.List[str] = []):
        super().__init__(parent)

        # setup the ui stuff
        self.setWidgetResizable(True)

        self._main_widget = QtWidgets.QWidget()
        self._layout = QFlowLayout(self._main_widget)

        self._tag_input = QtWidgets.QLineEdit()
        # self._tag_input.setPlaceholderText('Type in a new tag and hit enter...')
        self._tag_input.setFixedWidth(10)
        self._tag_input.setStyleSheet('border: 0px')
        self._tag_input.setContentsMargins(0, 5, 0, 0)
        self._tag_input.keyReleaseEvent = self.__tag_input_key_release_event

        self._layout.addWidget(self._tag_input)

        self._tag_input.palette().color(QtGui.QPalette.Background)
        tag_input_color = self._tag_input.palette().color(QtGui.QPalette.Background)
        self.setStyleSheet(f'background-color: rgb({tag_input_color.red()}, {tag_input_color.green()}, {tag_input_color.blue()})')

        self.setWidget(self._main_widget)

        # setup all other things
        self.__font_calculator = QtGui.QFontMetrics(QtWidgets.QApplication.font())
        self.__tags = {}
        self.__tag_suggestions = tag_suggestions
        self._tag_suggestions = False
        self._check_for_doubles = True

    def focusInEvent(self, a0: QtGui.QFocusEvent) -> None:
        """Sets the focus always to `self._tag_input`"""
        self._tag_input.setFocus()

    def addTag(self, text: str) -> bool:
        """
        Adds a new tag

        Args:
            text: The text of the new tag

        Returns:
            If the tag was added successfully
        """
        # if `self._check_for_doubles` is True, it checks if the tag already exists
        if self._check_for_doubles and text.lower() in (tag.lower() for tag in self.__tags.keys()):
            self.onDoubledTag(text)
            return False
        else:
            # a new tag
            tag = self.__QTagFrame(self, text)
            # tag.setStyleSheet('border: 0px; margin: 0px; padding: 0px')

            self.__tags[text] = tag
            for tag_name in self.__tag_suggestions:
                # if the tag is in `self.__tag_suggestions` it will be removed from there
                if tag_name.lower() == text:
                    self.__tag_suggestions.remove(tag_name)
                    self.enableTagSuggestions(self._tag_suggestions)
                    break
            # insert the tag before the line edit
            self._layout.addWidget(tag, -1)
            return True

    def clear(self, input=True) -> None:
        """
        Clears all tags

        Args:
            input: If True, the current text in the line edit will be cleared as well
        """
        for i in range(self._layout.count()):
            widget = self._layout.itemAt(i).widget()
            if type(widget) == QtWidgets.QLineEdit and input:
                widget.clear()
            elif type(widget) == self.__QTagFrame:
                self.removeTag(widget.text())

    def enableCheckForDoubles(self, check_for_doubles) -> None:
        """
        Enables if a new tag, when its going to be added, should be checked if it already exists

        Args:
            check_for_doubles: True if double checking should be active, False if not
        """
        self._check_for_doubles = check_for_doubles

    def enableTagSuggestions(self, tag_suggestions: bool) -> None:
        """
        Enables whenever a new tag is typed in that suggestions from `self.__tag_suggestions(...)` should be showing or not.
        They can be added on the class initialization or via `setTagSuggestions`

        Args:
            tag_suggestions: If tag suggestions should be active or not
        """
        self._tag_suggestions = tag_suggestions
        if tag_suggestions:
            # sets the completer for the line edit
            completer = QtWidgets.QCompleter(self.__tag_suggestions)
            completer.setCompletionMode(completer.InlineCompletion)
            completer.setCaseSensitivity(QtCore.Qt.CaseSensitive)
            self._tag_input.setCompleter(completer)
        else:
            self._tag_input.setCompleter(None)

    def onDoubledTag(self, text: str) -> None:
        """
        This method gets called if `self._check_for_doubles` is True (can be set via `enableCheckForDoubles(...)`)
        and a new tag which already exists are going to be added.
        This method is actually there to display an error message

        Args:
            text: The text of the new tag
        """
        button = QtWidgets.QMessageBox()

        # if `self.tag_input.keyReleaseEvent`is not overridden, it would trigger itself when the enter key is pressed
        # to close the popup and open it again. idk why
        def reset(a0: QtGui.QKeyEvent):
            if a0.key() in (QtCore.Qt.Key_Return, QtCore.Qt.Key_Enter):
                button.destroy()
                self._tag_input.keyReleaseEvent = self.__tag_input_key_release_event

        self._tag_input.keyReleaseEvent = reset

        # shows the warning
        button.warning(self, 'Tag already exists', f'The tag {text} already exists')

    def removeTag(self, tag: str) -> None:
        """
        Removes a tag

        Args:
            The tag to remove
        """
        if tag in self.__tags:
            self._layout.removeWidget(self.__tags[tag])
            del self.__tags[tag]

    def setTags(self, tags: typing.List[str]) -> None:
        """
        Replaces all current tags with tag from the `tags` argument

        Args:
            tags: The new tags to be set
        """
        self.clear()
        for tag in tags:
            self.addTag(tag)

    def setTagSuggestions(self, suggestions: typing.List[str]) -> None:
        """
        Sets the tag suggestions. They will be used if `self._tag_suggestions` is True (can be set via `enableTagSuggestions(...)`)
        and will be shown if a new tag is typed in

        Args:
            suggestions: The new tag suggestions
        """
        self.__tag_suggestions = suggestions
        self.enableTagSuggestions(self._tag_suggestions)

    def tags(self) -> typing.Union[typing.List[str]]:
        """
        Returns all tag names

        Returns:
            All tag names
        """
        return list(self.__tags.keys())

    def __tag_input_key_release_event(self, a0: QtGui.QKeyEvent) -> None:
        """
        The `keyReleaseEvent(...)` of the line edit. Whenever return / enter is pressed, the current text in the line edit
        will be added as new tag. It also expands the width of the line edit if the text in it is over an specific limit
        """
        if a0.key() in (QtCore.Qt.Key_Enter, QtCore.Qt.Key_Return):
            # adds the tag
            if self.addTag(self._tag_input.text()):
                self._tag_input.clear()
                self._tag_input.setFixedWidth(10)
            return
        elif a0.key() in (QtCore.Qt.Key_Backspace, QtCore.Qt.Key_Delete):
            # calculates the current line edit text width
            width = self.__font_calculator.width(self._tag_input.text())
            if width + 10 < self._tag_input.width():
                self._tag_input.setFixedWidth(width + 10)
        else:
            # calculates the current line edit text width
            width = self.__font_calculator.width(self._tag_input.text())
            # this resizes the tag input if the text in it will be longer than it's width.
            # not the best way, but it does what it does
            if width + 20 > self._tag_input.width():
                self._tag_input.setFixedWidth(width + 20)

    class __QTagFrame(QtWidgets.QFrame):
        """The tag class for the QTagEdit tags"""

        def __init__(self, parent, text: str):
            super().__init__()

            # setup the ui stuff
            self.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed))
            self.setLayout(QtWidgets.QHBoxLayout())
            self.setContentsMargins(3, 0, 3, 0)

            # the tag name label
            self._name = QtWidgets.QLabel()
            self._name.setText(text)
            self._name.setStyleSheet('background: transparent')

            # the tag delete button
            self._delete_button = QtWidgets.QPushButton()
            self._delete_button.setStyleSheet('QPushButton {'
                                              '     background: transparent;'
                                              '     border: 0px;'
                                              '}')
            self._delete_button.clicked.connect(self.onDeleteButtonClick)
            self._delete_button.setCursor(QtCore.Qt.PointingHandCursor)

            self.layout().addWidget(self._name)
            self.layout().addWidget(self._delete_button)

            # setup all other things
            self.__parent = parent
            self._text = text

        def onDeleteButtonClick(self) -> None:
            """This will get triggered if the delete button on a tag is pressed"""
            self.__parent.removeTag(self.text())
            self.__parent._tag_input.setFocus()

        def paintEvent(self, a0: QtGui.QPaintEvent) -> None:
            """Styles the tag"""
            dark_color = self.palette().color(QtGui.QPalette.Background).darker()
            light_color = self.palette().color(QtGui.QPalette.Background).lighter()

            painter = QtGui.QPainter(self)
            painter.setRenderHint(QtGui.QPainter.Antialiasing)
            painter.setPen(QtGui.QPen(dark_color, 1, QtCore.Qt.SolidLine))
            painter.setBrush(QtGui.QBrush(dark_color, QtCore.Qt.SolidPattern))

            # draws the tag 'filling'
            painter.drawRoundedRect(0, 0, self.width() - 5, self.height(), self.height() / 2, self.height() / 2)
            painter.setPen(QtGui.QPen(dark_color, 1, QtCore.Qt.SolidLine))
            painter.setBrush(QtGui.QBrush(light_color, QtCore.Qt.SolidPattern))
            painter.drawEllipse(QtCore.QPointF(self.width() - 25, self.height() / 2), 8, 8)
            painter.drawText(QtCore.QRectF(self.width() - 35, 0, 20, 30),
                             QtCore.Qt.AlignCenter,
                             '✕')

        def text(self) -> str:
            """
            Returns the current tag text

            Returns:
                The tag text

            """
            return self._text


"""
The animation objects for iconify
"""

from enum import Enum
from typing import Optional, Sequence, Tuple

from Custom_Widgets.iconify.qt import QtCore, QtGui


class GlobalTick(QtCore.QObject):
    """
    A singleton timer used to trigger all animation objects created by iconify
    """

    timeout = QtCore.Signal()

    _instance = None  # type: Optional[GlobalTick]

    def __init__(self):
        # type: () -> None
        # Note: No parent so it's owned by Qt
        super(GlobalTick, self).__init__()
        self._tick = QtCore.QTimer()
        self._tick.timeout.connect(self.timeout.emit)
        self._tick.setInterval(17)  # 60fps (ish)
        self._tick.start()

    @classmethod
    def instance(cls):
        # type: () -> GlobalTick
        """
        Return the global instance of the ticker.

        Returns
        -------
        GlobalTick
        """
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance


class BaseAnimation(QtCore.QObject):
    """
    The base class that should be used for all animations.
    """

    # Emitted when the animation steps
    tick = QtCore.Signal()

    _minFrame = 0
    _maxFrame = 100

    def __init__(self, parent=None):
        # type: (Optional[QtCore.QObject]) -> None
        super(BaseAnimation, self).__init__(parent=parent)
        self._frame = self._minFrame
        self._active = False

    def __add__(self, other):
        # type: (object) -> BaseAnimation
        if isinstance(other, BaseAnimation):
            concatAnim = _ConcatAnim()
            concatAnim.setAnimations((self, other))
            return concatAnim
        raise ValueError("Unsupported operation!")

    def transform(self, rect):
        # type: (QtCore.QRect) -> QtGui.QTransform
        """
        Return a QtGui.QTransform for the current frame that will be used
        when drawing an image in the provided QRect.

        Parameters
        ----------
        rect : QtCore.QRect

        Returns
        -------
        QtGui.QTransform
        """
        return QtGui.QTransform()

    def start(self):
        # type: () -> None
        """
        Start the animation.
        """
        GlobalTick.instance().timeout.connect(self._tick)
        self._active = True

    def stop(self):
        # type: () -> None
        """
        Stop the animation and reset the current frame back to the start.
        """
        self.pause()
        self._frame = self._minFrame

    def pause(self):
        # type: () -> None
        """
        Stop the animation and maintain the current frame
        """
        try:
            GlobalTick.instance().timeout.disconnect(self._tick)
        except RuntimeError:
            pass
        self._active = False

    def toggle(self):
        # type: () -> None
        """
        Start or stop the animation based on it's current state.
        """
        if self.active():
            self.pause()
        else:
            self.start()

    def active(self):
        # type: () -> bool
        """
        Indicate if the animation is currently playing.

        Returns
        -------
        bool
        """
        return self._active

    def frame(self):
        # type: () -> int
        """
        Return the current frame of the animation.

        Returns
        -------
        int
        """
        return self._frame

    def forceTick(self):
        # type: () -> None
        """
        Manually increment the current frame.
        """
        self._tick()

    def incrementFrame(self):
        # type: () -> None
        """
        Called when the animation is ticked allowing subclasses to provide
        custom frame step logic e.g. single shot animations.
        """
        if self._frame == self._maxFrame:
            self._frame = self._minFrame
        else:
            self._frame += 1

    def _tick(self):
        # type: () -> None
        self.incrementFrame()
        self.tick.emit()


class SingleShotMixin(object):
    """
    A mixin that overrides the incrementFrame logic so the
    animation loops once and then stops itself.
    """

    def incrementFrame(self):  # type: ignore[misc]  # noqa: F821
        # type: (BaseAnimation) -> None
        if self._frame == self._maxFrame:
            self._frame = self._minFrame
            self.stop()
        else:
            self._frame += 1


class Spin(BaseAnimation):
    """
    A simple spinning animation
    """

    _maxFrame = 59

    class Directions(Enum):

        CLOCKWISE = 0
        ANTI_CLOCKWISE = 1

    def __init__(self, direction=Directions.CLOCKWISE):
        # type: (Spin.Directions) -> None
        super(Spin, self).__init__()
        self._direction = direction

    def transform(self, size):
        # type: (QtCore.QSize) -> QtGui.QTransform
        halfSize = size / 2

        rotation = 6 if self._direction == Spin.Directions.CLOCKWISE else -6

        xfm = QtGui.QTransform()
        xfm = xfm.translate(halfSize.width(), halfSize.height())
        xfm = xfm.scale(0.8, 0.8)
        xfm = xfm.rotate(rotation * self._frame)
        xfm = xfm.translate(-halfSize.width(), -halfSize.height())

        return xfm


class SingleShotSpin(SingleShotMixin, Spin):
    """
    A single shot spinning animation
    """


class Breathe(BaseAnimation):

    _maxFrame = 100

    @staticmethod
    def _parametricEase(t):
        # type: (float) -> float
        sqt = t * t
        return sqt / (2.0 * (sqt - t) + 1.0)

    def transform(self, size):
        # type: (QtCore.QSize) -> QtGui.QTransform
        halfWay = self._maxFrame / 2

        if self._frame > halfWay:
            t = float(self._frame - halfWay) / halfWay
            m = self._parametricEase(t)
            scale = 0.9 - (0.2 * m)

        else:
            t = float(self._frame) / halfWay
            m = self._parametricEase(t)
            scale = 0.7 + (0.2 * m)

        halfSize = size / 2

        xfm = QtGui.QTransform()
        xfm = xfm.translate(halfSize.width(), halfSize.height())

        xfm = xfm.scale(scale, scale)
        xfm = xfm.translate(-halfSize.height(), -halfSize.width())

        return xfm


class _ConcatAnim(BaseAnimation):

    _maxFrame = 100000

    def __init__(self, parent=None):
        # type: (Optional[QtGui.QObject]) -> None
        super(_ConcatAnim, self).__init__(parent=parent)
        self._anims = ()  # type: Tuple[BaseAnimation, ...]

    def setAnimations(self, anims):
        # type: (Sequence[BaseAnimation]) -> None
        self._anims = tuple(anims)
        self._maxFrame = max([a._maxFrame for a in anims])

    def transform(self, rect):
        # type: (QtCore.QRect) -> QtGui.QTransform
        xfm = QtGui.QTransform()

        for anim in self._anims:
            xfm = anim.transform(rect) * xfm

        return xfm

    def start(self):
        # type: () -> None
        for anim in self._anims:
            anim.start()
        GlobalTick.instance().timeout.connect(self._tick)
        self._active = True

    def stop(self):
        # type: () -> None
        for anim in self._anims:
            anim.stop()
        self._frame = self._minFrame
        GlobalTick.instance().timeout.disconnect(self._tick)
        self._active = False

    def pause(self):
        # type: () -> None
        for anim in self._anims:
            anim.pause()
        GlobalTick.instance().timeout.disconnect(self._tick)
        self._active = False

    def toggle(self):
        # type: () -> None
        if self.active():
            self.pause()
        else:
            self.start()

    def forceTick(self):
        # type: () -> None
        for anim in self._anims:
            anim.forceTick()
        super(_ConcatAnim, self).forceTick()



from qtpy.QtCore import Qt, QMimeData, Signal
from qtpy.QtGui import QDrag, QPixmap, QPaintEvent, QPainter, QCursor
from qtpy.QtWidgets import QWidget, QLabel, QStyle, QStyleOption, QVBoxLayout

class DragTargetIndicator(QLabel):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setContentsMargins(25, 5, 25, 5)
        self.setStyleSheet(
            "QLabel {  border: 3px dotted; text-align: center}"
        )

        self.setText("Drop Here")


class QDragItem(QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Store data separately from display label, but use label for default.
        self.data = ""
        self.original_parent = None
        self.original_pos = None

    def mousePressEvent(self, e):
        if e.button() == Qt.LeftButton:
            self.original_parent = self.parentWidget()
            self.original_pos = self.pos()
            drag = QDrag(self)
            mime = QMimeData()
            drag.setMimeData(mime)

            pixmap = QPixmap(self.size())
            self.render(pixmap)
            drag.setPixmap(pixmap)
            self.setCursor(Qt.ClosedHandCursor)
            drag.exec_(Qt.MoveAction)


    def mouseMoveEvent(self, e):
        if e.buttons() == Qt.LeftButton:
            drag = QDrag(self)
            mime = QMimeData()
            drag.setMimeData(mime)

            pixmap = QPixmap(self.size())
            self.render(pixmap)
            drag.setPixmap(pixmap)
            self.setCursor(Qt.OpenHandCursor)
            drag.exec_(Qt.MoveAction)
            

    def mouseReleaseEvent(self, e):
        self.setCursor(Qt.ArrowCursor)

    def paintEvent(self, event: QPaintEvent):
        opt = QStyleOption()
        opt.initFrom(self)
        painter = QPainter(self)
        self.style().drawPrimitive(QStyle.PE_Widget, opt, painter, self)



class QDragWidget(QWidget):
    """
    Generic list sorting handler.
    """

    orderChanged = Signal(list)

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.setAcceptDrops(True)

        # Initialize layout as None initially
        # self.layout = None

        # Add the drag target indicator. This is invisible by default,
        # we show it and move it around while the drag is active.
        self._drag_target_indicator = DragTargetIndicator()
        self._drag_target_indicator.hide()


    def dragEnterEvent(self, e):
        e.accept()
        self.widgetLayout = self.layout()
        self.widget = e.source()

    def dragLeaveEvent(self, e):
        self._drag_target_indicator.hide()
        self.widget.show()
        e.ignore()
            

    def dragMoveEvent(self, e):
        # Find the correct location of the drop target, so we can move it there.
        index = self.findDropLocation(e)
        if index is not None:
            # Inserting moves the item if its already in the layout.
            self.widgetLayout.insertWidget(index, self._drag_target_indicator)
            # Hide the item being dragged.
            e.source().hide()
            # Show the target.
            self._drag_target_indicator.show()
            # Resize the target indicator to match the size of the widget
            self._drag_target_indicator.setMinimumSize(self.widget.size())
        e.accept()

    def dropEvent(self, e):
        widget = e.source()
        # Use drop target location for destination, then remove it.
        self._drag_target_indicator.hide()
        index = self.widgetLayout.indexOf(self._drag_target_indicator)
        
        # Get the position of the drop event relative to the parent DragWidget
        drop_position = self.mapFromGlobal(QCursor.pos())
        try:
            # Check if the drop event occurred within the bounds of the DragWidget
            if self.rect().contains(drop_position):
                if index == -1:  # Target indicator not found (dropping at the end)
                    index = self.widgetLayout.count()  # Append at the end of the layout
                
                # If index is greater than the length of the layout, append the widget at the end
                if index > self.widgetLayout.count():
                    index = self.widgetLayout.count()
                
                # If the drop was accepted, insert the widget at the drop location.
                self.widgetLayout.insertWidget(index, widget)
                self.orderChanged.emit(self.getItemData())
                widget.show()
                self.widgetLayout.activate()
                
                # Remove the drag target indicator from the layout.
                self.widgetLayout.removeWidget(self._drag_target_indicator)
                
                e.accept()
            else:
                # If the drop occurred outside the DragWidget, return the widget to its original position.
                widget.show()
                e.ignore()
            
            # Ensure the event is only processed once
            e.accept()
    
        except Exception as e:
            print(e)


    def findDropLocation(self, e):
        pos = e.pos()
        spacing = self.widgetLayout.spacing() / 2

        for n in range(self.widgetLayout.count()):
            # Get the widget at each index in turn.
            w = self.widgetLayout.itemAt(n).widget()

            if self.widgetLayout == QVBoxLayout:
                # Drag drop vertically.
                drop_here = (
                    pos.y() >= w.y() - spacing
                    and pos.y() <= w.y() + w.size().height() + spacing
                )
            else:
                # Drag drop horizontally.
                drop_here = (
                    pos.x() >= w.x() - spacing
                    and pos.x() <= w.x() + w.size().width() + spacing
                )

            if drop_here:
                # Drop over this target.
                break

        return n

    def getItemData(self):
        data = []
        for n in range(self.widgetLayout.count()):
            if self.widgetLayout.itemAt(n) is None:
                continue
            # Get the widget at each index in turn.
            w = self.widgetLayout.itemAt(n).widget()
            if hasattr(w, "data"):
                # The target indicator has no data.
                data.append(w.data)
        return data
    
    def paintEvent(self, event: QPaintEvent):
        opt = QStyleOption()
        opt.initFrom(self)
        painter = QPainter(self)
        self.style().drawPrimitive(QStyle.PE_Widget, opt, painter, self)


"""
Image location support
"""

import fnmatch
import os
from typing import List

from kids.cache import cache

_USER_DIR = os.path.expanduser('~/.iconify')
_ICON_PATH = os.environ.get('ICONIFY_PATH', _USER_DIR).split(os.pathsep)


class IconNotFoundError(Exception):
    pass


def addIconDirectory(directoryLocation):
    # type: (str) -> None
    """
    Add the provided path to the list of directories that iconify will use
    when looking for svg files.

    Parameters
    ----------
    directoryLocation : str
    """
    _ICON_PATH.append(directoryLocation)


@cache
def findIcon(iconPath):
    # type: (str) -> str
    """
    Use the provided string to find an svg file on the iconify path.

    Any `:`'s in the provided string will be replaced with the current
    platform's directory separator.

    Parameters
    ----------
    iconPath : str

    Returns
    -------
    str
    """
    if os.path.isabs(iconPath):
        if not os.path.isfile(iconPath):
            raise IconNotFoundError(
                "Unable to locate icon file: {}".format(iconPath)
            )
        return iconPath
    else:
        iconPath = iconPath.replace(":", os.sep)
        for dir_ in _ICON_PATH:
            absIconPath = os.path.join(dir_, iconPath + ".svg")
            if os.path.isfile(absIconPath):
                return absIconPath

        raise IconNotFoundError(
            "Unable to find an icon on the ICONIFY_PATH that matches '{}'".
            format(iconPath)
        )


def listIcons():
    # type: () -> List[str]
    """
    Return all icons found on the iconify path.

    Returns
    -------
    List[str]
    """
    matches = []  # type: List[str]
    for dir_ in _ICON_PATH:
        for root, dirnames, filenames in os.walk(dir_):
            for filename in fnmatch.filter(filenames, '*.svg'):
                filePath = os.path.join(root, filename)
                filePath = filePath.replace(dir_, "")
                filePath, _ = os.path.splitext(filePath)
                filePath = filePath.lstrip(os.sep).replace(os.sep, ":")
                matches.append(str(filePath))
    return matches



from qtpy.QtCore import Qt, Property, Signal
from qtpy.QtGui import QColor, QPainter, QFont
from qtpy.QtWidgets import QFrame, QSizePolicy
import os

class QBadgeWidget(QFrame):
    script_dir = os.path.dirname(os.path.realpath(__file__))
    WIDGET_ICON = os.path.join(script_dir, "components/icons/badge.png")
    WIDGET_TOOLTIP = "A custom badge widget"
    WIDGET_DOM_XML = """
    <ui language='c++'>
        <widget class='QBadgeWidget' name='customBadge'>
        </widget>
    </ui>
    """
    WIDGET_MODULE="Custom_Widgets.QBadgeWidget"

    clicked = Signal()

    def __init__(
            self, 
            parent=None,
            text="Badge text",
            background_color=QColor(255, 0, 0),
            text_color=QColor(255, 255, 255)
            ):
        QFrame.__init__(self, parent=parent)
        
        self._text = text
        self._background_color = background_color
        self._text_color = text_color
        
        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        # self.setMinimumSize(80, 30)
        self.setCursor(Qt.PointingHandCursor)
        self.setGeometry(0, 0, 80, 30)

    
    @Property(str)
    def text(self):
        return self._text
    
    @text.setter
    def text(self, value):
        self._text = value
        self.update()
    
    @Property(QColor)
    def backgroundColor(self):
        return self._background_color
    
    @backgroundColor.setter
    def backgroundColor(self, color):
        self._background_color = color
        self.update()
    
    @Property(QColor)
    def textColor(self):
        return self._text_color
    
    @textColor.setter
    def textColor(self, color):
        self._text_color = color
        self.update()
    
    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        
        painter.setBrush(self._background_color)
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(self.rect(), 10, 10)
        
        painter.setPen(self._text_color)
        painter.setFont(QFont('Arial', 10))
        painter.drawText(self.rect(), Qt.AlignCenter, self._text)
    
    def mousePressEvent(self, event):
        self.clicked.emit()

    def showEvent(self, event):
        super().showEvent(event)
        self.update()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'info.ui'
##
## Created by: Qt User Interface Compiler version 6.6.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from qtpy.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from qtpy.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from qtpy.QtWidgets import (QApplication, QFrame, QHBoxLayout, QLabel,
    QPushButton, QSizePolicy, QVBoxLayout, QWidget)

class Ui_Form(object):
    def setupUi(self, Form):
        if not Form.objectName():
            Form.setObjectName(u"Form")
        Form.resize(444, 79)
        self.verticalLayout = QVBoxLayout(Form)
        self.verticalLayout.setSpacing(5)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(5, 5, 5, 5)
        self.frame = QFrame(Form)
        self.frame.setObjectName(u"frame")
        self.frame.setFrameShape(QFrame.StyledPanel)
        self.frame.setFrameShadow(QFrame.Raised)
        self.horizontalLayout = QHBoxLayout(self.frame)
        self.horizontalLayout.setSpacing(5)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(5, 5, 5, 5)
        self.iconlabel = QLabel(self.frame)
        self.iconlabel.setObjectName(u"iconlabel")
        self.iconlabel.setMinimumSize(QSize(20, 20))
        self.iconlabel.setMaximumSize(QSize(20, 20))
        self.iconlabel.setScaledContents(True)

        self.horizontalLayout.addWidget(self.iconlabel, 0, Qt.AlignLeft)

        self.titlelabel = QLabel(self.frame)
        self.titlelabel.setObjectName(u"titlelabel")
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.titlelabel.sizePolicy().hasHeightForWidth())
        self.titlelabel.setSizePolicy(sizePolicy)
        font = QFont()
        font.setBold(True)
        self.titlelabel.setFont(font)
        self.titlelabel.setWordWrap(True)

        self.horizontalLayout.addWidget(self.titlelabel)

        self.closeButton = QPushButton(self.frame)
        self.closeButton.setObjectName(u"closeButton")
        icon = QIcon(QIcon.fromTheme(u"application-exit"))
        self.closeButton.setIcon(icon)

        self.horizontalLayout.addWidget(self.closeButton, 0, Qt.AlignRight)


        self.verticalLayout.addWidget(self.frame, 0, Qt.AlignTop)

        self.frame_2 = QFrame(Form)
        self.frame_2.setObjectName(u"frame_2")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.frame_2.sizePolicy().hasHeightForWidth())
        self.frame_2.setSizePolicy(sizePolicy1)
        self.frame_2.setFrameShape(QFrame.StyledPanel)
        self.frame_2.setFrameShadow(QFrame.Raised)
        self.verticalLayout_2 = QVBoxLayout(self.frame_2)
        self.verticalLayout_2.setSpacing(5)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(5, 5, 5, 5)
        self.bodyLabel = QLabel(self.frame_2)
        self.bodyLabel.setObjectName(u"bodyLabel")
        self.bodyLabel.setWordWrap(True)

        self.verticalLayout_2.addWidget(self.bodyLabel)


        self.verticalLayout.addWidget(self.frame_2)


        self.retranslateUi(Form)

        QMetaObject.connectSlotsByName(Form)
    # setupUi

    def retranslateUi(self, Form):
        Form.setWindowTitle(QCoreApplication.translate("Form", u"Form", None))
        self.iconlabel.setText("")
        self.titlelabel.setText(QCoreApplication.translate("Form", u"Title", None))
        self.closeButton.setText("")
        self.bodyLabel.setText(QCoreApplication.translate("Form", u"Body", None))
    # retranslateUi




from qtpy.QtCore import Qt, QEvent, QPropertyAnimation, QEasingCurve, Property, Signal, QSize, QPoint, QTimer, QCoreApplication
from qtpy.QtWidgets import QPushButton, QWidget, QLabel, QStyleOption, QStyle, QHBoxLayout, QVBoxLayout, QSizePolicy, QGraphicsOpacityEffect, QSpacerItem, QApplication, QGraphicsDropShadowEffect
from qtpy.QtGui import QPainter, QIcon, QPaintEvent, QEnterEvent, QMouseEvent, QHoverEvent, QColor, QCursor
import os

# Import your custom sidebar and utility functions
from Custom_Widgets.QCustomSidebar import QCustomSidebar 
from Custom_Widgets.Utils import replace_url_prefix, is_in_designer, get_icon_path
from Custom_Widgets.Log import *

class QCustomSidebarButton(QPushButton):
    clicked = Signal()

    # Define XML for Qt Designer
    script_dir = os.path.dirname(os.path.realpath(__file__))
    WIDGET_ICON = os.path.join(script_dir, "components/icons/arrow_forward.png")
    WIDGET_TOOLTIP = "A custom button that interacts with the sidebar"
    WIDGET_DOM_XML = """
    <ui language='c++'>
        <widget class='QCustomSidebarButton' name='customSidebarButton'>
        </widget>
    </ui>
    """
    WIDGET_MODULE = "Custom_Widgets.QCustomSidebarButton"

    def __init__(self, parent=None, *args):
        super().__init__(parent)

        # Install event filter for the whole application
        app = QCoreApplication.instance()
        if app is not None:
            app.installEventFilter(self)
        elif self.parent():
            self.parent().installEventFilter(self)

        # Store original text and icon for resetting
        self.original_text = ""
        self.original_icon = self.icon()
        self._label_hidden = False
        self._hideOnCollapse = True
        self._text_prefix_spaces = 5

        self._fading_out = False

        self._is_hovered = False
        self._floating_widget = None
        self._hover_timer = QTimer(self)
        self._hover_timer.setSingleShot(True)
        self._hover_timer.timeout.connect(self._show_floating_button)
        self._float_position = None


    @Property(bool)
    def hideOnCollapse(self):
        """Whether to hide this label when the sidebar collapses."""
        return self._hideOnCollapse

    @hideOnCollapse.setter
    def hideOnCollapse(self, hide):
        self._hideOnCollapse = hide

    @Property(int)
    def textPrefixSpaces(self):
        """Get number of spaces to prepend to the text."""
        return self._text_prefix_spaces

    @textPrefixSpaces.setter
    def textPrefixSpaces(self, num_spaces):
        """Set number of spaces to prepend to the text."""
        self._text_prefix_spaces = num_spaces

        self.update()

    # Define the property for labelHidden state
    @Property(bool)
    def labelHidden(self, designable = False):
        return self._label_hidden

    @labelHidden.setter
    def labelHidden(self, state):
        self._label_hidden = state
        self.style().unpolish(self)  # Refresh style
        self.style().polish(self)
        self.update()

    @Property(str, designable=True)
    def labelText(self):
        """Returns the label text for the button (read-only)."""
        return self.original_text or ""

    @labelText.setter
    def labelText(self, text):
        """Sets the original label text for the button."""
        self.original_text = text
        self.update()

    @property
    def text(self):
        return super().text()

    @text.setter
    def text(self, value):
        super().setText(value or 'Sidebar Button')

    def paintEvent(self, event: QPaintEvent):
        """Custom paint event to draw the button with opacity."""
        opt = QStyleOption()
        opt.initFrom(self)
        painter = QPainter(self)
        self.style().drawControl(QStyle.CE_PushButton, opt, painter, self)

        if self.original_text and not self.labelHidden:  
            self.setText(self.original_text) 

        elif self.labelHidden:
            self.setText("**clear")

        self.update()
        
        super().paintEvent(event) 

    def setText(self, text):
        """Override setText to store the raw text and apply the prefix spaces."""
        if text == "**clear":
            super().setText("")

        else:

            if self.original_text != text:
                self.labelText = text
            super().setText(self.getPrefixedText(text))

    def update(self):
        if self.original_text and not self.labelHidden:  
            self.setText(self.original_text) 

        elif self.labelHidden:
            self.setText("**clear")

        super().update()


    def getPrefixedText(self, text):
        return " " * self._text_prefix_spaces + text.lstrip()

    def connect_to_parent(self):
        """Connect to the closest QCustomSidebar parent if necessary."""
        self.parent_sidebar = self.parent()  # Start with the direct parent
        while self.parent_sidebar and not isinstance(self.parent_sidebar, QCustomSidebar):
            self.parent_sidebar = self.parent_sidebar.parent()  # Move up the hierarchy

        if self.parent_sidebar:
            self.parent_sidebar.onCollapsed.connect(self.hideButtonLabel)
            self.parent_sidebar.onExpanded.connect(self.showButtonLabel)

            self.parent_sidebar.onCollapsing.connect(self.showButtonLabel)
            self.parent_sidebar.onExpanding.connect(self.showButtonLabel)

            if self.parent_sidebar and self.parent_sidebar.isCollapsed():
                self.hideButtonLabel()
            else:
                self.showButtonLabel()

    def hideButtonLabel(self):
        """Hide the button label by clearing the text."""
        if not self.original_text:
            self.original_text = self.text()
        
        if self.original_text:
            self.setText("**clear")  # Clear the button text
            self.labelHidden = True

        # Set the custom property for labelHidden state
        self.labelHidden = True

    def hideButtonIcon(self):
        """Hide the button icon by setting it to an empty QIcon."""
        self.original_icon = self.icon()
        self.setIcon(QIcon())  # Set an empty icon

    def showButtonLabel(self):
        """Show the button label by restoring the original text."""
        if self.original_text:  # Check if there is original text to show
            self.setText(self.original_text)  # Restore the original text
            self.labelHidden = False

        # Unset the custom property for labelHidden state
        self.labelHidden = False

        self._fade_out_floating_button()


    def showButtonIcon(self):
        """Show the button icon by restoring the original icon."""
        if not self.original_icon.isNull():  # Check if there is an original icon to show
            self.setIcon(self.original_icon)  # Restore the original icon

    def showEvent(self, e):
        super().showEvent(e)
        self.connect_to_parent()
        # Adjust size and update the widget
        # self.adjustSize()
        self.update()

        try:
            if self.parent_sidebar and self.parent_sidebar.isCollapsed():
                if not self.labelHidden:
                    self.hideButtonLabel()
            if self.parent_sidebar and self.parent_sidebar.isExpanded():
                if self.labelHidden:
                    self.showButtonLabel()
        except Exception as e:
            logException(e)

    def enterEvent(self, event: QEnterEvent):
        """Show the button label when the button is hovered, even if the sidebar is collapsed."""
        #FIXME: self.parent_sidebar.isCollapsed() and self.parent_sidebar.isExpanded() do not always return the expected bool
        # print(self.parent_sidebar.isCollapsed(), self.parent_sidebar.isExpanded())
        if self.parent_sidebar and (self.parent_sidebar.isCollapsed() or not self.parent_sidebar.isExpanded()):
            # Start a timer to avoid immediate hover effect
            self._hover_timer.start(2000)
        super().enterEvent(event)

    def leaveEvent(self, event):
        """Handle mouse leave event to set _is_hovered to False."""
        # self._fade_out_floating_button()
        super().leaveEvent(event)

    def _delete_floating_button(self, e):
        """Hide the button label when the hover ends, return to original collapsed state."""
        if self._floating_widget:
            self._fade_out_floating_button()  # Fade out the floating button
        self._hover_timer.stop()  # Cancel any hover event in progress

    def _show_floating_button(self):
        """Show the floating button only if the mouse is still over the main button."""
        if not self.labelHidden:
            return
        # Check if the mouse is still over the main button
        if not self.rect().contains(self.mapFromGlobal(QCursor.pos())):
            return  # Mouse is no longer hovering over the button, so don't show the floating button

        # If mouse is still over the button, create and display the floating button
        if not self._floating_widget:
            self._create_floating_button()

            self._floating_widget.show()
            self._floating_widget.setMinimumSize(self._floating_widget.sizeHint())
            self._floating_button.adjustSize()
            self._floating_widget.adjustSize()
            self._floating_widget.move(self._calculate_floating_position())


    def _create_floating_button(self):
        """Create the floating version of the button."""
        # Create a QWidget as the container
        self._floating_widget = QWidget(self)
        self._floating_widget.setObjectName("floatingButtonWidget") #for css styling
        # self._floating_widget.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        
        # Create the QPushButton
        self._floating_button = QCustomSidebarButton(" " * self._text_prefix_spaces + self.original_text, self._floating_widget)
        self._floating_widget.hideOnCollapse = False
        self._floating_button.setIcon(self.icon())
        self._floating_button.setObjectName(self.objectName())

        # Create the shadow effect
        shadow = QGraphicsDropShadowEffect(self._floating_button)
        shadow.setBlurRadius(10)  # Set the blur radius for the shadow
        shadow.setColor(QColor(0, 0, 0, 160))  # Set the shadow color (can be customized)
        shadow.setOffset(0, 0)  # Set the offset for the shadow (horizontal, vertical)

        # Apply the shadow effect to the widget
        self._floating_button.setGraphicsEffect(shadow)
        
        # Create a QVBoxLayout
        layout = QVBoxLayout(self._floating_widget)
        layout.addWidget(self._floating_button)
        
        # Set the layout margins and spacing to zero
        layout.setContentsMargins(10, 10, 10, 10)
        layout.setSpacing(0)

        # Set the widget's layout
        self._floating_widget.setLayout(layout)
        # Raise the widget to the front
        self._floating_widget.raise_()
        self._floating_widget.setAttribute(Qt.WA_TranslucentBackground, True)
        self._floating_widget.setWindowFlags(Qt.FramelessWindowHint | Qt.ToolTip | Qt.Popup)

        # 
        self._floating_button.showEvent = self._fade_in_floating_button

        # Connect events from the floating button to the main button's event handlers
        self._floating_button.mousePressEvent = self.mousePressEvent
        self._floating_button.mouseReleaseEvent = self.mouseReleaseEvent
        self._floating_button.enterEvent = self.enterEvent
        self._floating_button.leaveEvent = self.leaveEvent

        self._fading_out = False

    def _pass_event_to_main_button(self, event):
        """Pass all events from the floating button to the main button."""
        # Forward the event to the main button
        return super(QPushButton, self._floating_button).event(event)

    def _fade_in_floating_button(self, e = None):
        """Fade in the floating button."""
        if self._floating_widget and not self._fading_out:
            # fade animation
            self._opacityAni = QPropertyAnimation(self._floating_widget, b'windowOpacity', self._floating_widget)
            self._opacityAni.setDuration(500)
            self._opacityAni.setStartValue(0)
            self._opacityAni.setEndValue(1)
            self._opacityAni.start()

    def _fade_out_floating_button(self):
        """Fade out the floating button."""
        if self._fading_out:
            return

        if self._floating_widget:
            self._fading_out = True
            # fade animation
            self._opacityAni = QPropertyAnimation(self._floating_widget, b'windowOpacity', self._floating_widget)
            self._opacityAni.setDuration(500)
            self._opacityAni.setStartValue(1)
            self._opacityAni.setEndValue(0)
            self._opacityAni.finished.connect(self._hide_floating_button)
            self._opacityAni.start()

    def _hide_floating_button(self):
        """Hide the floating button after the fade-out."""
        if self._floating_widget:
            self._floating_widget.hide()  # Hide the button
            self._floating_widget.deleteLater()  # Schedule for deletion
            self._floating_widget = None  # Clear reference

    def _calculate_floating_position(self):
        """Calculate the exact relative position for the floating button."""
        # Get the position of the main button relative to its parent 
        floating_button_pos = self.mapToGlobal(QPoint(-10, -10))
        return floating_button_pos

    def resizeEvent(self, event):
        """Update floating button position on window resize."""
        if self._floating_widget:
            self._floating_widget.move(self._calculate_floating_position())
        super().resizeEvent(event)

    def moveEvent(self, event):
        """Update floating button position on window move."""
        if self._floating_widget:
            self._floating_widget.move(self._calculate_floating_position())
        super().moveEvent(event)

    def eventFilter(self, obj, event: QEvent):
        if event.type() in (QEvent.MouseButtonPress, QEvent.MouseButtonRelease, QEvent.MouseButtonDblClick, QEvent.MouseMove):
            # Handle the mouse event here
            
            local_pos = self.mapFromGlobal(event.globalPos())
            if hasattr(self, "_floating_widget") and self._floating_widget and not self._floating_button.rect().contains(local_pos):
                self._fade_out_floating_button()

        return super().eventFilter(obj, event)



import os
import sys
import re
import shutil
import subprocess
import qtpy
from qtpy.QtDesigner import QDesignerFormWindowInterface
from qtpy.QtGui import QIcon
from qtpy.QtCore import QSettings

# Import custom logging module
from Custom_Widgets.Log import *

def get_absolute_path(relative_path):
    """Convert a relative path to an absolute path based on the script's directory."""
    import __main__
    
    # Try multiple methods to find the main script directory
    possible_dirs = []
    
    # Method 1: __main__.__file__
    if hasattr(__main__, '__file__'):
        possible_dirs.append(os.path.dirname(os.path.abspath(__main__.__file__)))
    
    # Method 2: sys.argv[0] (your original approach)
    if sys.argv[0]:
        possible_dirs.append(os.path.dirname(os.path.abspath(sys.argv[0])))
    
    # Method 3: Current working directory
    possible_dirs.append(os.path.abspath(os.getcwd()))
    
    # Use the first valid directory that contains the relative path when joined
    for main_dir in possible_dirs:
        test_path = os.path.join(main_dir, relative_path)
        if os.path.exists(test_path):
            return os.path.normpath(test_path)
    
    # If no existing path found, use the first method and hope for the best
    main_dir = possible_dirs[0] if possible_dirs else os.path.abspath(os.getcwd())
    return os.path.normpath(os.path.join(main_dir, relative_path))

def replace_url_prefix(url, new_prefix):
    pattern = re.compile(r':/[^/]+/')
    return pattern.sub( new_prefix + '/', url, 1)

def get_icon_path(icon: QIcon | str) -> str:
    """Return the correct path for a themed icon. Handle both QIcon and string paths."""
    settings = QSettings()
    # Check if the 'ICONS-COLOR' setting is defined
    if settings.value("ICONS-COLOR") is not None:
        # Get the normal color and derive the icon folder name from it
        normal_color = settings.value("ICONS-COLOR")
        icons_folder = normal_color.replace("#", "")  # Strip the '#' for folder naming

        # Regular expression to remove the old prefix in the icon path
        prefix_to_remove = re.compile(r'^Qss/icons/[^/]+/')

        if isinstance(icon, QIcon):
            # Handle QIcon by converting to the appropriate icon path
            icon_url = icon.name()  # Assuming the QIcon has a name or can be represented
            return icon.addFile(re.sub(prefix_to_remove, f'Qss/icons/{icons_folder}/', replace_url_prefix(icon_url, "Qss/icons")))

        elif isinstance(icon, str):
            # If the input is a string (file path), process it directly
            return re.sub(prefix_to_remove, f'Qss/icons/{icons_folder}/', replace_url_prefix(icon, "Qss/icons"))

    return icon 

def is_in_designer(self):
    """Check if the widget is in Qt Designer."""
    # logInfo(QDesignerFormWindowInterface.findChild(self))
    return QDesignerFormWindowInterface.findFormWindow(self) is not None


def createQrcFile(contents, filePath):
    # Ensure the directory for the filePath exists
    os.makedirs(os.path.dirname(filePath), exist_ok=True)
    
    # Save QRC content to a file
    with open(filePath, 'w', encoding='utf-8') as qrc_file:
        qrc_file.write(contents)

    # print(f'QRC file generated: {filePath}')

def qrcToPy(qrcFile, pyFile):
    """
    Convert a Qt Resource Collection (qrc) file to a Python file.

    Parameters:
    - qrc_file (str): Path to the input qrc file.
    - py_file (str): Path to the output py file.
    """
    try:
        if qtpy.API_NAME == "PyQt5":
            rcc_command = 'pyrcc5'
        elif qtpy.API_NAME == "PyQt6":
            rcc_command = 'pyrcc6'
        elif qtpy.API_NAME == "PySide2":
            rcc_command = 'pyside2-rcc'
        elif qtpy.API_NAME == "PySide6":
            rcc_command = 'pyside6-rcc'
        else:
            raise Exception("Error: Unknown QT binding/API Name", qtpy.API_NAME)

        print(f'{rcc_command} "{qrcFile}" -o "{pyFile}"')
        subprocess.run(f'{rcc_command} "{qrcFile}" -o "{pyFile}"')
        
    except Exception as e:
        print("Error converting qrc to py:", e)

def uiToPy(uiFile, pyFile):
    """
    Convert a Qt UI file to a Python file.

    Parameters:
    - uiFile (str): Path to the input UI file.
    - pyFile (str): Path to the output Python file.
    """
    try:
        if qtpy.API_NAME == "PyQt5":
            pyuic_command = 'pyuic5'
        elif qtpy.API_NAME == "PyQt6":
            pyuic_command = 'pyuic6'
        elif qtpy.API_NAME == "PySide2":
            pyuic_command = 'pyside2-uic'
        elif qtpy.API_NAME == "PySide6":
            pyuic_command = 'pyside6-uic'
        else:
            raise Exception("Error: Unknown Qt binding/API Name", qtpy.API_NAME)

        os.system(f'{pyuic_command} "{uiFile}" -o "{pyFile}"')

    except Exception as e:
        print("Error converting ui to py:", e)

def renameFolder(old_name, new_name):
    try:
        # Check if the destination directory exists
        if os.path.exists(new_name):
            # Remove the destination directory if it exists
            shutil.rmtree(new_name)

        # Rename the folder
        os.rename(old_name, new_name)
    except Exception as e:
        pass


class SharedData:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(SharedData, cls).__new__(cls)
            cls._instance.file_urls = []  # Initialize an empty list for file URLs
        return cls._instance

    def add_file_url(self, file_url):
        """Add a new file URL to the list."""
        if file_url not in self.file_urls:  # Prevent duplicates
            self.file_urls.append(file_url)

    def get_file_urls(self):
        """Return the list of file URLs."""
        return self.file_urls

    def clear_file_urls(self):
        """Clear the list of file URLs."""
        self.file_urls.clear()

    def url_exists(self, file_url):
        """Check if a file URL exists in the list."""
        return file_url in self.file_urls




## SPINN DESIGN CODE
# YOUTUBE: (SPINN TV) https://www.youtube.com/spinnTv
# WEBSITE: spinncode.com
import os
import sys
import __main__


from qtpy.QtCore import QCoreApplication, Qt, QSettings, QPoint, QSize, Signal, QEvent
from qtpy.QtGui import QCursor, QPaintEvent,QColor, QMouseEvent, QPainter, QIcon
from qtpy.QtWidgets import QPushButton, QLabel, QTabWidget, QCheckBox, QMainWindow, QWidget, QVBoxLayout, QStyle, QStyleOption, QGraphicsDropShadowEffect, QToolBox

import re

from Custom_Widgets.JSonStyles import loadJsonStyle
from Custom_Widgets.QCustomTheme import QCustomTheme
from Custom_Widgets.Utils import is_in_designer, SharedData

from Custom_Widgets.Log import *
from Custom_Widgets.QCustomComponentLoader import QCustomComponentLoader
from Custom_Widgets.QCustomHamburgerMenu import QCustomHamburgerMenu
from Custom_Widgets.FileMonitor import QSsFileMonitor

script_dir = os.path.dirname(os.path.abspath(sys.argv[0])).replace("\\", "/")

class QMainWindow(QMainWindow):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.clickPosition = None  
        self.normalGeometry = self.geometry()
        self.iconsWorker = None
        self.allIconsWorker = None
        
        self.orginazationName = ""
        self.applicationName = ""
        self.orginazationDomain = ""

        self._border_width = 5 #for resizing
        self.borderWidth = self._border_width
        self._is_resizing = False
        self._margin_applied = False
        self._shadow_applied = False

        self.themeEngine =  QCustomTheme()
        self.themeEngine.onThemeChanged.connect(self.applyThemeIcons)
        self._init_theme_name = self.themeEngine.theme

        self.themeEngine.sassCompilationProgress = self.sassCompilationProgress

        self.shared_data = SharedData()
        self.qss_watcher = None

        self._win_restored = False
        self._loaded_side_drawers = set()
        self._loaded_side_drawer_files = set()
        self._custom_side_drawers = ""

        self.customSideDrawers = ""

        self._qss_file_monitor = QSsFileMonitor.instance()
        self.qss_watcher = None
        self.liveCompileQss = False
        self._qss_file_monitor.start_qss_file_listener(self.themeEngine)

    def saveGeometryToSettings(self):
        """Save the current window geometry (position and size) to settings."""
        self.settings = QSettings()
        self.settings.setValue("window/size", self.size())
        self.settings.setValue("window/position", self.pos())
        self.settings.setValue("window/state", self.isMaximized())


    def loadGeometryFromSettings(self):
        """Load and apply the saved window geometry (position and size) from settings."""
        # Get saved size and position
        if self._win_restored:
            return
        try:
            self._win_restored = True

            self.settings = QSettings()
            size = self.settings.value("window/size")  # Default size if not saved
            position = self.settings.value("window/position")  # Default position if not saved
            state = self.settings.value("window/state")  

            # Apply saved geometry
            # self.resize(size)
            # self.move(position)

            # if state:
            #     self.showMaximized()

        except:
            self._win_restored = False

    def sassCompilationProgress(self, n):
        pass

    @property
    def themes(self):
        self.themeEngine =  QCustomTheme()
        return self.themeEngine.themes

    def applyThemeIcons(self):
        if hasattr(self, "ui"):
            try:
                # Check the module name where ui is loaded from
                self.ui_module_name = self.ui.__module__.split('.')[-1]

                # Replace "ui_" with empty string only at the start
                if self.ui_module_name.startswith("ui_"):
                    self.ui_module_name = self.ui_module_name[len("ui_"):]

            except Exception as e:
                self.ui_module_name = ""
                logException(e)

            try:
                self.themeEngine.applyIcons(self.ui, ui_file_name=self.ui_module_name)

            except Exception as e:
                logException(e, "Error loading theme icons for : "+ str(self.ui))
                
            self.restyleAllButtonGroups()

        # Also update side drawers if any
        self.reloadCustomSideDrawersThemeIcons()

    # Update restore button icon on maximizing or minimizing window
    def updateRestoreButtonIcon(self):
        normal_color = self.themeEngine.iconsColor
        icons_folder = normal_color.replace("#", "")

        prefix_to_remove = re.compile(r'^Qss/icons/[^/]+/')
        self.maximizedIcon = re.sub(prefix_to_remove, 'Qss/icons/'+icons_folder+'/', self.maximizedIcon)
        self.normalIcon = re.sub(prefix_to_remove, 'Qss/icons/'+icons_folder+'/', self.normalIcon)

        # If window is maxmized
        if self.isMaximized():
            # Change Iconload
            if len(str(self.maximizedIcon)) > 0:
                # self.restoreBtn.setIcon(QtGui.QIcon(str(self.maximizedIcon)))
                self.restoreBtn.setNewIcon(str(self.maximizedIcon))
        else:
            # Change Icon
            if len(str(self.normalIcon)) > 0:
                # self.restoreBtn.setIcon(QtGui.QIcon(str(self.normalIcon)))
                self.restoreBtn.setNewIcon(str(self.normalIcon))

    def restore_or_maximize_window(self):
        self.toggleWindowSize("")

    # Function to Move window on mouse drag event on the tittle bar
    def moveWindow(self, e):
        # Detect if the window is  normal size
        if not self.isMaximized(): #Not maximized
            # Move window only when window is normal size
            #if left mouse button is clicked (Only accept left mouse button clicks)
            if e.buttons() == Qt.LeftButton:
                #Move window
                if self.clickPosition is not None:
                    self.move(self.pos() + e.globalPos() - self.clickPosition)
                    self.clickPosition = e.globalPos()
                    e.accept()

            self.normalGeometry = self.geometry()
        # else:
        #     self.showNormal()

        self.saveGeometryToSettings()

    def toggleWindowSize(self, e):
        if self.isMaximized():
            self.showNormal()
        else:
            self.showMaximized()

        self.updateRestoreButtonIcon()
        self.updateProperties()

    def updateProperties(self):
        if self.isMaximized():
            self.setProperty("window-state", "maximized")
        else:
            self.setProperty("window-state", "normal")

        self.style().unpolish(self) 
        self.style().polish(self) 

    ## Check Button Groups
    def checkButtonGroup(self, button = None):
        if self.sender() is not None:
            btn = self.sender()
        else:
            btn = button
                
        group = btn.group
        groupBtns = getattr(self, "group_btns_"+str(group))
        active = getattr(self, "group_active_"+str(group))
        notActive = getattr(self, "group_not_active_"+str(group))

        for x in groupBtns:
            if x == btn and self.sender() is not None:
                x.setStyleSheet(self.themeEngine.styleVariablesFromTheme(active))
                x.active = True

            elif  x.active and self.sender() is None:
                x.setStyleSheet(self.themeEngine.styleVariablesFromTheme(active))
                x.active = True

            else:
                x.setStyleSheet(self.themeEngine.styleVariablesFromTheme(notActive))
                x.active = False

    def restyleAllButtonGroups(self):
        """
        Loop through all button groups and reapply active/inactive styles based on the theme.
        """
        # Iterate over each group number, assuming groups are counted as integers from 1
        grp_count = 1
        while hasattr(self, f"group_btns_{grp_count}"):
            group_buttons = getattr(self, f"group_btns_{grp_count}")
            active_style = getattr(self, f"group_active_{grp_count}", "")
            not_active_style = getattr(self, f"group_not_active_{grp_count}", "")

            # Apply styles for each button based on its active state
            for btn in group_buttons:
                if btn.active:
                    btn.setStyleSheet(self.themeEngine.styleVariablesFromTheme(active_style))
                else:
                    btn.setStyleSheet(self.themeEngine.styleVariablesFromTheme(not_active_style))
            
            grp_count += 1  # Move to the next group
    
    def loadCustomSideDrawers(self):
        """Load QCustomHamburgerMenu from customSideDrawers list."""
        _custom_side_drawers = [f.strip() for f in self.customSideDrawers.split(",") if f.strip()]
        
        if is_in_designer(self):
            return
        
        for drawer_file in _custom_side_drawers:
            if drawer_file in self._loaded_side_drawer_files:
                continue
                
            try:
                # Create component loader
                component_loader = QCustomComponentLoader(self)
                component_loader.previewComponent = True
                component_loader.loadComponent(filePath=drawer_file)

                component_loader.setMaximumSize(QSize(0, 0))
                # remove margins
                component_loader.layout().setContentsMargins(0, 0, 0, 0)
                # Add the loader widget to layout
                self.centralWidget().layout().addWidget(component_loader)

                # Store reference
                if component_loader.ui:
                    component_loader.form = component_loader.ui
                    component_loader.shownForm = component_loader.ui
                    component_loader.component = component_loader.ui
                
                # Search for QCustomHamburgerMenu in the component loader
                hamburger_menu = self._findHamburgerMenu(component_loader)
                if hamburger_menu:
                    self._reparentHamburgerMenu(hamburger_menu)
                    self._loaded_side_drawers.add(component_loader)
                else:
                    logWarning(f"No QCustomHamburgerMenu found in loaded side drawers, file: {drawer_file}, container: {component_loader.objectName()}")
                
                self._loaded_side_drawer_files.add(drawer_file)
                
                component_loader.hide()

            except Exception as e:
                logError(f"Failed to load side drawer {drawer_file}: {str(e)}")
                    
    def reloadCustomSideDrawersThemeIcons(self):
        """Reload theme icons for all loaded side drawers."""
        for drawer in self._loaded_side_drawers:
            try:
                # drawer.themeEngine = self.themeEngine
                drawer.applyThemeIcons()
            except Exception as e:
                logError(f"Failed to reload theme icons for side drawer {drawer.objectName()}: {str(e)}")
                
    def _findHamburgerMenu(self, root_widget):
        """
        Recursively search for QCustomHamburgerMenu widget.
        Returns the first top-level hamburger menu found.
        """
        # Check if root widget itself is a hamburger menu
        if isinstance(root_widget, QCustomHamburgerMenu):
            return root_widget
        
        # Search through all child widgets recursively
        def search_children(widget):
            for child in widget.children():
                if isinstance(child, QCustomHamburgerMenu):
                    return child
                # Recursively search deeper
                result = search_children(child)
                if result:
                    return result
            return None
        
        return search_children(root_widget)

    def getHamburgerMenu(self, name: str) -> QCustomHamburgerMenu:
        try:
            # Validate input
            if not name or not isinstance(name, str):
                raise ValueError("Hamburger menu name must be a non-empty string")
            
            # Search in central widget first
            hamburger = self.centralWidget().findChild(QCustomHamburgerMenu, name)
            
            # If not found, search recursively through all widgets
            if not hamburger:
                hamburger = self.findChild(QCustomHamburgerMenu, name)
            
            if not hamburger:
                logWarning(f"Hamburger menu '{name}' not found")
                return None
                
            return hamburger
            
        except ValueError as e:
            logError(f"Invalid hamburger menu name: {e}")
            raise
        except Exception as e:
            logError(f"Unexpected error while finding hamburger menu '{name}': {str(e)}")
            return None
    
    def getHamburgerWidget(self, menu_name: str, widget_name: str) -> QWidget:
        hamburger = self.getHamburgerMenu(menu_name)
        if hamburger:
            return hamburger.getWidget(widget_name)
        return None

    
    def _reparentHamburgerMenu(self, hamburger_menu):
        """
        Reparent hamburger menu to central widget and remove from any layout.
        """
        hamburger_menu.hide()
        # Store current geometry and properties
        old_geometry = hamburger_menu.geometry()
        old_parent = hamburger_menu.parent()
        # Remove from any existing layout
        # if old_parent and old_parent.layout():
        #     old_parent.layout().removeWidget(hamburger_menu)
        
        # # Reparent to central widget
        hamburger_menu.setParent(self.centralWidget())
        
        # Restore geometry and ensure it's visible
        hamburger_menu.setGeometry(old_geometry)
        hamburger_menu.raise_()
        hamburger_menu.show()
    

    def showEvent(self, e: QEvent):
        super().showEvent(e)
        self.updateProperties()
        self.applyThemeIcons()
        try:
            self.borderWidth = self._border_width - self.shadowBlurRadius * 2.2
        except:
            pass

        self.applyBorderRadius()
        self.restyleAllButtonGroups()
        
        self.loadCustomSideDrawers()


    def resizeEvent(self, e: QEvent):
        self.saveGeometryToSettings()
        super().resizeEvent(e)
        # self.applyBorderRadius()
        self.update()

    def moveEvent(self, e: QEvent):
        self.applyBorderRadius()
        super().moveEvent(e)
        self.saveGeometryToSettings()
    
    def paintEvent(self, e: QPaintEvent):
        super().paintEvent(e)
        opt = QStyleOption()
        opt.initFrom(self)
        painter = QPainter(self)
        self.style().drawPrimitive(QStyle.PE_Widget, opt, painter, self)

        # self.applyBorderRadius()

        centralWidget = self.centralWidget()
        centralWidget.setGeometry(
            (self.width() - centralWidget.width()) // 2,
            (self.height() - centralWidget.height()) // 2,
            centralWidget.width(),
            centralWidget.height()
        )

        if self._init_theme_name != self.themeEngine.theme:
            self._init_theme_name = self.themeEngine.theme
            self.applyThemeIcons()
            self.loadGeometryFromSettings() #assumed its being shown for the first time

        
    # Override the changeEvent method
    def changeEvent(self, event):
        # Check if the event is a window state change
        if event.type() == QEvent.WindowStateChange:
            self.applyBorderRadius()
            self.saveGeometryToSettings()

        # Call the base class implementation
        super().changeEvent(event)

    def applyBorderRadius(self):
        if self.isFrameless():
            try:
                if self.centralWidget():
                    centralWidget = self.centralWidget()
                    layout = centralWidget.layout()
                    name = centralWidget.objectName()

                    # self.layout().setAlignment(centralWidget, Qt.AlignCenter)
                    
                    if self.isMaximized():
                        self.layout().setContentsMargins(0, 0, 0, 0)
                        layout.setContentsMargins(0, 0, 0, 0)

                        centralWidget.setMaximumSize(self.width(), self.height())

                        self._margin_applied = False

                        self.centralWidget().setGraphicsEffect(None)

                    else:
                        centralWidget.setMaximumSize(self.width() - self.shadowBlurRadius * 2.2, self.height() - self.shadowBlurRadius * 2.2)
                        if not self._margin_applied:
                            top = self.height() - self.shadowBlurRadius * 2.2
                            left = self.width() - centralWidget.width() // 2

                            self.layout().setContentsMargins(40, 40, 40, 40)
                            layout.setContentsMargins(self.borderRadius * .7, self.borderRadius * .7, self.borderRadius * .7, self.borderRadius * .7)

                            self._margin_applied = True

                        if not self._shadow_applied:
                            self.applyDropShadow()
                            self._shadow_applied = True

            except Exception as e:
                print(e)

        self.update()

    def applyDropShadow(self):
        self.centralWidget().setGraphicsEffect(None)
        self.shadow = QGraphicsDropShadowEffect(self)
        if self.shadowBlurRadius > 0:
            self.shadow.setColor(QColor(self.shadowColor))
            self.shadow.setBlurRadius(self.shadowBlurRadius)
            self.shadow.setXOffset(self.shadowXOffset)
            self.shadow.setYOffset(self.shadowYOffset)

            ## # Appy shadow to central widget
            self.centralWidget().setGraphicsEffect(self.shadow)

        self.update()

    def mousePressEvent(self, event: QMouseEvent):
        # Get the current position of the mouse
        self.clickPosition = event.globalPos()
        # We will use this value to move the window
        # Hide floating widgets
        cursor = QCursor()
        xPos = cursor.pos().x()
        yPos = cursor.pos().y()
        # if hasattr(self, "floatingWidgets"):
        #     for x in self.floatingWidgets:
        #         if hasattr(x, "autoHide") and x.autoHide:
        #             x.collapseMenu()

        if not self.isFrameless():
            return
        
        if event.button() == Qt.LeftButton:
            self._drag_position = event.globalPosition().toPoint()
            self._resize_direction = self._detect_resize_direction(event.position().toPoint())
            if self._resize_direction:
                self._is_resizing = True

    def mouseMoveEvent(self, event: QMouseEvent):
        if not self.isFrameless():
            return

        if self.isMaximized():
            self.borderWidth = self._border_width
        else:
            self.borderWidth = self._border_width + self.shadowBlurRadius
        
        self._update_cursor(event.position().toPoint())

        if self._is_resizing:
            self._resize_window(event)
            
    def mouseReleaseEvent(self, event: QMouseEvent):
        if event.button() == Qt.LeftButton:
            self._is_resizing = False
            self._resize_direction = None
            self.unsetCursor()

    def _detect_resize_direction(self, pos: QPoint):
        rect = self.rect()
        if pos.x() < self.borderWidth:
            if pos.y() < self.borderWidth:
                return "top_left"
            elif pos.y() > rect.height() - self.borderWidth:
                return "bottom_left"
            else:
                return "left"
        elif pos.x() > rect.width() - self.borderWidth:
            if pos.y() < self.borderWidth:
                return "top_right"
            elif pos.y() > rect.height() - self.borderWidth:
                return "bottom_right"
            else:
                return "right"
        elif pos.y() < self.borderWidth:
            return "top"
        elif pos.y() > rect.height() - self.borderWidth:
            return "bottom"
        else:
            return None

    def _update_cursor(self, pos: QPoint):
        direction = self._detect_resize_direction(pos)
        if direction in ["left", "right"]:
            self.setCursor(Qt.SizeHorCursor)
        elif direction in ["top", "bottom"]:
            self.setCursor(Qt.SizeVerCursor)
        elif direction in ["top_left", "bottom_right"]:
            self.setCursor(Qt.SizeFDiagCursor)
        elif direction in ["top_right", "bottom_left"]:
            self.setCursor(Qt.SizeBDiagCursor)
        else:
            self.unsetCursor()

    def _resize_window(self, event: QMouseEvent):
        # self._update_cursor(event.position().toPoint())
        if self._resize_direction:
            rect = self.geometry()
            delta = event.globalPosition().toPoint() - self._drag_position
            if self._resize_direction == "left":
                rect.setLeft(rect.left() + delta.x())
            elif self._resize_direction == "right":
                rect.setRight(rect.right() + delta.x())
            elif self._resize_direction == "top":
                rect.setTop(rect.top() + delta.y())
            elif self._resize_direction == "bottom":
                rect.setBottom(rect.bottom() + delta.y())
            elif self._resize_direction == "top_left":
                rect.setTopLeft(rect.topLeft() + delta)
            elif self._resize_direction == "top_right":
                rect.setTopRight(rect.topRight() + delta)
            elif self._resize_direction == "bottom_left":
                rect.setBottomLeft(rect.bottomLeft() + delta)
            elif self._resize_direction == "bottom_right":
                rect.setBottomRight(rect.bottomRight() + delta)

            self.setGeometry(rect)
            self._drag_position = event.globalPosition().toPoint()
        
    def isFrameless(window: QMainWindow) -> bool:
        return bool(window.windowFlags() & Qt.FramelessWindowHint)

def mouseReleaseEvent(self, QMouseEvent):
    cursor = QCursor()
    # self.ui.frame.setGeometry(QRect(cursor.pos().x(), cursor.pos().y(), 151, 111))





########################################################################
## SPINN DESIGN CODE
# YOUTUBE: (SPINN TV) https://www.youtube.com/spinnTv
# WEBSITE: spinncode.com
########################################################################

########################################################################
## IMPORTS
########################################################################
import os

########################################################################
## MODULE UPDATED TO USE QT.PY
########################################################################
from qtpy.QtCore import QCoreApplication, QSettings

from Custom_Widgets.Log import *

########################################################################
## QT APP SETTINGS
########################################################################
class QAppSettings():
    def __init__(self, parent=None):
        super(QAppSettings, self).__init__(parent)
        ########################################################################
        ## CREATE APP SETTINGS
        ########################################################################

    def updateAppSettings(self, generateIcons: bool = True, reloadJson: bool = True, paintEntireApp: bool = True, QtDesignerMode: bool = False):
        
        if not hasattr(self, "themeEngine"):
            return

        themeEngine = self.themeEngine

        if len(str(themeEngine.orginazationName)) > 0:
            QCoreApplication.setOrganizationName(str(themeEngine.orginazationName))
        if len(str(themeEngine.applicationName)) > 0:
            QCoreApplication.setApplicationName(str(themeEngine.applicationName))
        if len(str(themeEngine.orginazationDomain)) > 0:
            QCoreApplication.setOrganizationDomain(str(themeEngine.orginazationDomain))

        settings = QSettings()

        # if theme not set
        init_theme_set = settings.value("INIT-THEME-SET")
        if settings.value("THEME") is None or not init_theme_set:
            for theme in themeEngine.themes:
                if theme.defaultTheme:
                    # update app theme
                    if (init_theme_set is None or not init_theme_set):
                        settings.setValue("THEME", theme.name)
                        settings.setValue("INIT-THEME-SET", True)
                        logInfo("Initial theme set...", theme.name)

        settings.setValue("THEMES-LIST", themeEngine.themes)

        #######################################################################
        # APPLY COMPILED STYLESHEET
        #######################################################################
        if reloadJson:
            themeEngine.reloadJsonStyles(update = False)
            
        themeEngine.applyCompiledSass(generateIcons = generateIcons, paintEntireApp = paintEntireApp)


########################################################################
## END
########################################################################


# Custom Widgets Registration for Qt Designer
# Author: Khamisi Kibet

# Import custom logging module
from Custom_Widgets.Log import *

# Ensure the logger is set up
setupLogger(designer = True)

import qtpy.QtDesigner as QtDesigner

logInfo("Registering Custom Widgets")

from Custom_Widgets.QCustomQMainWindow import QCustomQMainWindow

# Registering QCustomQMainWindow with error handling
try:
    logInfo("Registering QCustomQMainWindow")
    QtDesigner.QPyDesignerCustomWidgetCollection.registerCustomWidget(
        QCustomQMainWindow, module=QCustomQMainWindow.WIDGET_MODULE,
        tool_tip=QCustomQMainWindow.WIDGET_TOOLTIP, 
        xml=QCustomQMainWindow.WIDGET_DOM_XML,
        icon=QCustomQMainWindow.WIDGET_ICON, container=True, group="Main Window"
    )
except Exception as e:
    logException(e, message="Error registering QCustomQMainWindow")


from Custom_Widgets.QAvatarWidget import QAvatarWidget 

# Registering QAvatarWidget with error handling
try:
    logInfo("Registering QAvatarWidget")
    QtDesigner.QPyDesignerCustomWidgetCollection.registerCustomWidget(
        QAvatarWidget, module=QAvatarWidget.WIDGET_MODULE,
        tool_tip=QAvatarWidget.WIDGET_TOOLTIP, 
        xml=QAvatarWidget.WIDGET_DOM_XML,
        icon=QAvatarWidget.WIDGET_ICON
    )
except Exception as e:
    logException(e, message="Error registering QAvatarWidget")


from Custom_Widgets.QBadgeWidget import QBadgeWidget 

# Registering QBadgeWidget with error handling
try:
    logInfo("Registering QBadgeWidget")
    QtDesigner.QPyDesignerCustomWidgetCollection.registerCustomWidget(
        QBadgeWidget, module=QBadgeWidget.WIDGET_MODULE,
        tool_tip=QBadgeWidget.WIDGET_TOOLTIP, 
        xml=QBadgeWidget.WIDGET_DOM_XML,
        icon=QBadgeWidget.WIDGET_ICON
    )
except Exception as e:
    logException(e, message="Error registering QBadgeWidget")


from Custom_Widgets.AnalogGaugeWidget import AnalogGaugeWidget 

# Registering AnalogGaugeWidget with error handling
try:
    logInfo("Registering AnalogGaugeWidget")
    QtDesigner.QPyDesignerCustomWidgetCollection.registerCustomWidget(
        AnalogGaugeWidget, module=AnalogGaugeWidget.WIDGET_MODULE,
        tool_tip=AnalogGaugeWidget.WIDGET_TOOLTIP, 
        xml=AnalogGaugeWidget.WIDGET_DOM_XML,
        icon=AnalogGaugeWidget.WIDGET_ICON
    )
except Exception as e:
    logException(e, message="Error registering AnalogGaugeWidget")


from Custom_Widgets.QCustomThemeList import QCustomThemeList 

# Registering QCustomThemeList with error handling
try:
    logInfo("Registering QCustomThemeList")
    QtDesigner.QPyDesignerCustomWidgetCollection.registerCustomWidget(
        QCustomThemeList, module=QCustomThemeList.WIDGET_MODULE,
        tool_tip=QCustomThemeList.WIDGET_TOOLTIP, 
        xml=QCustomThemeList.WIDGET_DOM_XML,
        icon=QCustomThemeList.WIDGET_ICON
    )
except Exception as e:
    logException(e, message="Error registering QCustomThemeList")

from Custom_Widgets.QCustomThemeDarkLightToggle import QCustomThemeDarkLightToggle 

# Registering QCustomThemeDarkLightToggle with error handling
try:
    logInfo("Registering QCustomThemeDarkLightToggle")
    QtDesigner.QPyDesignerCustomWidgetCollection.registerCustomWidget(
        QCustomThemeDarkLightToggle, module=QCustomThemeDarkLightToggle.WIDGET_MODULE,
        tool_tip=QCustomThemeDarkLightToggle.WIDGET_TOOLTIP, 
        xml=QCustomThemeDarkLightToggle.WIDGET_DOM_XML,
        icon=QCustomThemeDarkLightToggle.WIDGET_ICON
    )
except Exception as e:
    logException(e, message="Error registering QCustomThemeDarkLightToggle")


from Custom_Widgets.QCustomCheckBox import QCustomCheckBox 

# Registering QCustomCheckBox with error handling
try:
    logInfo("Registering QCustomCheckBox")
    QtDesigner.QPyDesignerCustomWidgetCollection.registerCustomWidget(
        QCustomCheckBox, module=QCustomCheckBox.WIDGET_MODULE,
        tool_tip=QCustomCheckBox.WIDGET_TOOLTIP, 
        xml=QCustomCheckBox.WIDGET_DOM_XML,
        icon=QCustomCheckBox.WIDGET_ICON
    )
except Exception as e:
    logException(e, message="Error registering QCustomCheckBox")


from Custom_Widgets.QCustomSidebar import QCustomSidebar 

# Registering QCustomSidebar with error handling
try:
    logInfo("Registering QCustomSidebar")
    QtDesigner.QPyDesignerCustomWidgetCollection.registerCustomWidget(
        QCustomSidebar, module=QCustomSidebar.WIDGET_MODULE,
        tool_tip=QCustomSidebar.WIDGET_TOOLTIP, 
        xml=QCustomSidebar.WIDGET_DOM_XML,
        icon=QCustomSidebar.WIDGET_ICON, container=True, group="Sidebar"
    )
except Exception as e:
    logException(e, message="Error registering QCustomSidebar")


# ADD HAMBURGER MENU WIDGETS HERE - RIGHT AFTER SIDEBAR REGISTRATION

from Custom_Widgets.QCustomHamburgerMenu import QCustomHamburgerMenu

# Registering QCustomHamburgerMenu with error handling
try:
    logInfo("Registering QCustomHamburgerMenu")
    QtDesigner.QPyDesignerCustomWidgetCollection.registerCustomWidget(
        QCustomHamburgerMenu, module=QCustomHamburgerMenu.WIDGET_MODULE,
        tool_tip=QCustomHamburgerMenu.WIDGET_TOOLTIP, 
        xml=QCustomHamburgerMenu.WIDGET_DOM_XML,
        icon=QCustomHamburgerMenu.WIDGET_ICON, container=True, group="Hamburger Menu"
    )
except Exception as e:
    logException(e, message="Error registering QCustomHamburgerMenu")


from Custom_Widgets.QCustomHorizontalSeparator import QCustomHorizontalSeparator 

# Registering QCustomHorizontalSeparator with error handling
try:
    logInfo("Registering QCustomHorizontalSeparator")
    QtDesigner.QPyDesignerCustomWidgetCollection.registerCustomWidget(
        QCustomHorizontalSeparator, module=QCustomHorizontalSeparator.WIDGET_MODULE,
        tool_tip=QCustomHorizontalSeparator.WIDGET_TOOLTIP, 
        xml=QCustomHorizontalSeparator.WIDGET_DOM_XML,
        icon=QCustomHorizontalSeparator.WIDGET_ICON, container=False, group="Sidebar"
    )
except Exception as e:
    logException(e, message="Error registering QCustomHorizontalSeparator")

from Custom_Widgets.QCustomVerticalSeparator import QCustomVerticalSeparator 

# Registering QCustomVerticalSeparator with error handling
try:
    logInfo("Registering QCustomVerticalSeparator")
    QtDesigner.QPyDesignerCustomWidgetCollection.registerCustomWidget(
        QCustomVerticalSeparator, module=QCustomVerticalSeparator.WIDGET_MODULE,
        tool_tip=QCustomVerticalSeparator.WIDGET_TOOLTIP, 
        xml=QCustomVerticalSeparator.WIDGET_DOM_XML,
        icon=QCustomVerticalSeparator.WIDGET_ICON, container=False, group="Sidebar"
    )
except Exception as e:
    logException(e, message="Error registering QCustomVerticalSeparator")


from Custom_Widgets.QCustomSidebarLabel import QCustomSidebarLabel 

# Registering QCustomSidebarLabel with error handling
try:
    logInfo("Registering QCustomSidebarLabel")
    QtDesigner.QPyDesignerCustomWidgetCollection.registerCustomWidget(
        QCustomSidebarLabel, module=QCustomSidebarLabel.WIDGET_MODULE,
        tool_tip=QCustomSidebarLabel.WIDGET_TOOLTIP, 
        xml=QCustomSidebarLabel.WIDGET_DOM_XML,
        icon=QCustomSidebarLabel.WIDGET_ICON, group="Sidebar"
    )
except Exception as e:
    logException(e, message="Error registering QCustomSidebarLabel")


from Custom_Widgets.QCustomSidebarButton import QCustomSidebarButton 

# Registering QCustomSidebarButton with error handling
try:
    logInfo("Registering QCustomSidebarButton")
    QtDesigner.QPyDesignerCustomWidgetCollection.registerCustomWidget(
        QCustomSidebarButton, module=QCustomSidebarButton.WIDGET_MODULE,
        tool_tip=QCustomSidebarButton.WIDGET_TOOLTIP, 
        xml=QCustomSidebarButton.WIDGET_DOM_XML,
        icon=QCustomSidebarButton.WIDGET_ICON, group="Sidebar"
    )
except Exception as e:
    logException(e, message="Error registering QCustomSidebarButton")


from Custom_Widgets.QCustomProgressBars import QCustomRoundProgressBar 

# Registering QCustomRoundProgressBar with error handling
try:
    logInfo("Registering QCustomRoundProgressBar")
    QtDesigner.QPyDesignerCustomWidgetCollection.registerCustomWidget(
        QCustomRoundProgressBar, module=QCustomRoundProgressBar.WIDGET_MODULE,
        tool_tip=QCustomRoundProgressBar.WIDGET_TOOLTIP, 
        xml=QCustomRoundProgressBar.WIDGET_DOM_XML,
        icon=QCustomRoundProgressBar.WIDGET_ICON, group="Progressbars"
    )
except Exception as e:
    logException(e, message="Error registering QCustomRoundProgressBar")

from Custom_Widgets.QCustomComponent import QCustomComponent

# Registering QCustomComponent with error handling
try:
    logInfo("Registering QCustomComponent")
    QtDesigner.QPyDesignerCustomWidgetCollection.registerCustomWidget(
        QCustomComponent, module=QCustomComponent.WIDGET_MODULE,
        tool_tip=QCustomComponent.WIDGET_TOOLTIP, 
        xml=QCustomComponent.WIDGET_DOM_XML,
        icon=QCustomComponent.WIDGET_ICON, container=True, group="Component Container"
    )
except Exception as e:
    logException(e, message="Error registering QCustomComponent")

from Custom_Widgets.QCustomComponentContainer import QCustomComponentContainer

# Registering QCustomComponentContainer with error handling
try:
    logInfo("Registering QCustomComponentContainer")
    QtDesigner.QPyDesignerCustomWidgetCollection.registerCustomWidget(
        QCustomComponentContainer, module=QCustomComponentContainer.WIDGET_MODULE,
        tool_tip=QCustomComponentContainer.WIDGET_TOOLTIP, 
        xml=QCustomComponentContainer.WIDGET_DOM_XML,
        icon=QCustomComponentContainer.WIDGET_ICON, container=False, group="Component Container"
    )
except Exception as e:
    logException(e, message="Error registering QCustomComponentContainer")

from Custom_Widgets.QCustomQStackedWidget import QCustomQStackedWidget

try:
    logInfo("Registering QCustomQStackedWidget")
    QtDesigner.QPyDesignerCustomWidgetCollection.registerCustomWidget(
        QCustomQStackedWidget, module=QCustomQStackedWidget.WIDGET_MODULE,
        tool_tip=QCustomQStackedWidget.WIDGET_TOOLTIP, 
        xml=QCustomQStackedWidget.WIDGET_DOM_XML,
        icon=QCustomQStackedWidget.WIDGET_ICON, container=True
    )
except Exception as e:
    logException(e, message="Error registering QCustomQStackedWidget")

from Custom_Widgets.QCustomLoadingIndicators import QCustomQProgressBar 

# Registering QCustomQProgressBar with error handling
try:
    logInfo("Registering QCustomQProgressBar")
    QtDesigner.QPyDesignerCustomWidgetCollection.registerCustomWidget(
        QCustomQProgressBar, module=QCustomQProgressBar.WIDGET_MODULE,
        tool_tip=QCustomQProgressBar.WIDGET_TOOLTIP, 
        xml=QCustomQProgressBar.WIDGET_DOM_XML,
        icon=QCustomQProgressBar.WIDGET_ICON, group="Progressbars"
    )
except Exception as e:
    logException(e, message="Error registering QCustomQProgressBar")

from Custom_Widgets.QCustomQRGenerator import QCustomQRGenerator

# Registering QCustomQRGenerator with error handling
try:
    logInfo("Registering QCustomQRGenerator")
    QtDesigner.QPyDesignerCustomWidgetCollection.registerCustomWidget(
        QCustomQRGenerator, module=QCustomQRGenerator.WIDGET_MODULE,
        tool_tip=QCustomQRGenerator.WIDGET_TOOLTIP, 
        xml=QCustomQRGenerator.WIDGET_DOM_XML,
        icon=QCustomQRGenerator.WIDGET_ICON, container=True, group="QR Generator"
    )
except Exception as e:
    logException(e, message="Error registering QCustomQRGenerator")
    



import logging
import os
import sys
import traceback
from logging.handlers import RotatingFileHandler
from qtpy.QtCore import QSettings
from Custom_Widgets.Utils import is_in_designer

# Rich for beautiful console output
try:
    from rich.console import Console
    from rich.logging import RichHandler
    from rich.traceback import Traceback
    from rich.panel import Panel
    from rich.text import Text
    from rich.theme import Theme
    from rich import print as rprint
    RICH_AVAILABLE = True
except ImportError:
    RICH_AVAILABLE = False
    print("Rich not available. Install with: pip install rich")

# Custom theme for rich console
if RICH_AVAILABLE:
    custom_theme = Theme({
        "info": "cyan",
        "warning": "yellow",
        "error": "red",
        "critical": "bold red",
        "debug": "dim",
        "success": "green",
        "file": "blue",
        "folder": "magenta",
        "monitor": "bold blue"
    })
    console = Console(theme=custom_theme)

# Setup logger
def setupLogger(self=None, designer=False):
    logFilePath = os.path.join(os.getcwd(), "logs/custom_widgets.log")
    if designer or (self is not None and is_in_designer(self)):
        logFilePath = os.path.join(os.getcwd(), "logs/custom_widgets_designer.log")
    
    # Ensure the log directory exists
    logDirectory = os.path.dirname(logFilePath)
    if logDirectory != "" and not os.path.exists(logDirectory):
        os.makedirs(logDirectory)

    # Get the root logger
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    
    # Clear any existing handlers
    for handler in logger.handlers[:]:
        logger.removeHandler(handler)

    # Set up the rotating file handler
    logFileMaxSize = 10 * 1024 * 1024  # 10 MB
    backupCount = 5  # Keep up to 5 backup log files
    
    file_handler = RotatingFileHandler(
        logFilePath, 
        maxBytes=logFileMaxSize, 
        backupCount=backupCount,
        encoding='utf-8'
    )
    file_handler.setLevel(logging.DEBUG)
    file_formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )
    file_handler.setFormatter(file_formatter)
    logger.addHandler(file_handler)

    # Rich console handler for beautiful output
    if RICH_AVAILABLE:
        rich_handler = RichHandler(
            console=console,
            rich_tracebacks=True,
            tracebacks_show_locals=True,
            show_time=True,
            show_level=True,
            show_path=True
        )
        rich_handler.setLevel(logging.INFO)
    else:
        # Fallback to basic console handler
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.INFO)
        console_formatter = logging.Formatter(
            '%(asctime)s - %(levelname)s - %(message)s'
        )
        console_handler.setFormatter(console_formatter)
        logger.addHandler(console_handler)

# Retrieve QSettings
def get_show_custom_widgets_logs():
    settings = QSettings()
    return settings.value("showCustomWidgetsLogs", True, type=bool)

def set_show_custom_widgets_logs(value: bool):
    settings = QSettings()
    settings.setValue("showCustomWidgetsLogs", value)

# Enhanced logging functions with rich formatting
def logDebug(message, **kwargs):
    logging.debug(message, extra=kwargs)
    if get_show_custom_widgets_logs() and RICH_AVAILABLE:
        console.print(f"🔍 [debug]DEBUG:[/debug] {message}", **kwargs)

def logInfo(message, **kwargs):
    logging.info(message, extra=kwargs)
    if get_show_custom_widgets_logs():
        if RICH_AVAILABLE:
            console.print(f"ℹ️ [info]INFO:[/info] {message}", **kwargs)
        else:
            print(f"INFO: {message}")

def logWarning(message, **kwargs):
    logging.warning(message, extra=kwargs)
    if get_show_custom_widgets_logs():
        if RICH_AVAILABLE:
            console.print(f"⚠️ [warning]WARNING:[/warning] {message}", **kwargs)
        else:
            print(f"WARNING: {message}")

def logError(message, **kwargs):
    logging.error(message, extra=kwargs)
    if get_show_custom_widgets_logs():
        if RICH_AVAILABLE:
            console.print(f"❌ [error]ERROR:[/error] {message}", **kwargs)
        else:
            print(f"ERROR: {message}")

def logCritical(message, **kwargs):
    logging.critical(message, extra=kwargs)
    if get_show_custom_widgets_logs():
        if RICH_AVAILABLE:
            console.print(f"💥 [critical]CRITICAL:[/critical] {message}", **kwargs)
        else:
            print(f"CRITICAL: {message}")

def logSuccess(message, **kwargs):
    logging.info(f"SUCCESS: {message}", extra=kwargs)
    if get_show_custom_widgets_logs() and RICH_AVAILABLE:
        console.print(f"✅ [success]SUCCESS:[/success] {message}", **kwargs)

def logException(exception, message="Exception", **kwargs):
    logging.exception(f"{message}: {exception}", extra=kwargs)
    if get_show_custom_widgets_logs():
        if RICH_AVAILABLE:
            console.print(f"🚨 [error]EXCEPTION:[/error] {message}: {exception}", **kwargs)
            console.print(Traceback.from_exception(type(exception), exception, exception.__traceback__))
        else:
            print(f"EXCEPTION: {message}: {exception}")
            traceback.print_exc()

# File monitoring specific logging functions
def logFileMonitorStart(files_count, folder_path=None):
    message = f"Starting file monitor - Tracking {files_count} files"
    if folder_path:
        message += f" in [folder]{folder_path}[/folder]"
    logInfo(message)
    if RICH_AVAILABLE and get_show_custom_widgets_logs():
        console.print(Panel.fit(
            f"📁 [monitor]FILE MONITOR STARTED[/monitor]\n"
            f"• Files: [file]{files_count}[/file]\n"
            f"• Folder: [folder]{folder_path or 'N/A'}[/folder]",
            border_style="blue"
        ))

def logFileChange(path, action="modified"):
    logInfo(f"File [file]{path}[/file] has been {action}")
    if RICH_AVAILABLE and get_show_custom_widgets_logs():
        console.print(f"📝 [info]File {action}:[/info] [file]{os.path.basename(path)}[/file]")

def logFileListUpdate(new_files, removed_files, total_files):
    changes = []
    if new_files:
        changes.append(f"[success]+{len(new_files)} new[/success]")
    if removed_files:
        changes.append(f"[error]-{len(removed_files)} removed[/error]")
    
    if changes:
        logInfo(f"File list updated: {', '.join(changes)} - Total: {total_files} files")
        if RICH_AVAILABLE and get_show_custom_widgets_logs():
            console.print(f"🔄 [info]File list updated:[/info] {', '.join(changes)} - Total: [file]{total_files}[/file] files")

def logFileConversionStart(file_path):
    logInfo(f"Starting conversion of [file]{file_path}[/file]")
    if RICH_AVAILABLE and get_show_custom_widgets_logs():
        console.print(f"🛠️ [info]Converting:[/info] [file]{os.path.basename(file_path)}[/file]")

def logFileConversionComplete(file_path):
    logSuccess(f"Completed conversion of [file]{file_path}[/file]")
    if RICH_AVAILABLE and get_show_custom_widgets_logs():
        console.print(f"✅ [success]Converted:[/success] [file]{os.path.basename(file_path)}[/file]")

def logWidgetProcessing(widget_class, widget_name, details=""):
    details_text = f" - {details}" if details else ""
    logDebug(f"Processing widget: {widget_class} '{widget_name}'{details_text}")
    if RICH_AVAILABLE and get_show_custom_widgets_logs():
        console.print(f"🎛️ [debug]Widget:[/debug] {widget_class} '[file]{widget_name}[/file]'[dim]{details_text}[/dim]")

def logIconProcessing(widget_name, icon_url, widget_type="Widget"):
    logDebug(f"{widget_type} '{widget_name}' icon: {icon_url}")
    if RICH_AVAILABLE and get_show_custom_widgets_logs():
        console.print(f"🖼️ [debug]{widget_type} icon:[/debug] '[file]{widget_name}[/file]' → [file]{icon_url}[/file]")

def logJSONUpdate(json_file, data_summary):
    logInfo(f"Updated JSON file: {json_file} with {data_summary}")
    if RICH_AVAILABLE and get_show_custom_widgets_logs():
        console.print(f"📊 [info]JSON updated:[/info] [file]{json_file}[/file] - {data_summary}")

# Handle unhandled exceptions with rich formatting
def handle_unhandled_exception(exc_type, exc_value, exc_traceback):
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return
    
    formatted_traceback = ''.join(traceback.format_exception(exc_type, exc_value, exc_traceback))
    logging.critical("Unhandled exception occurred:\n%s", formatted_traceback)
    
    if get_show_custom_widgets_logs():
        if RICH_AVAILABLE:
            console.print(Panel.fit(
                f"[critical]UNHANDLED EXCEPTION[/critical]\n"
                f"[error]{exc_type.__name__}: {exc_value}[/error]",
                border_style="red"
            ))
            console.print(Traceback.from_exception(exc_type, exc_value, exc_traceback))
        else:
            print("UNHANDLED EXCEPTION:")
            print(formatted_traceback)

# Set the exception hook for unhandled exceptions
sys.excepthook = handle_unhandled_exception

# Initialize logger when module is imported
# setupLogger()